import os
import numpy as np
from .Abstract_Reranker import Abstract_Reranker
from tqdm import tqdm,trange
import ot
import cvxpy as cp

r"""
TaxRank, an algorithm to transfer the item-side fair re-ranking phase as a taxation process.
The TaxRank is designed for item-side fairness, its performance may be not good in group-side

TaxRank is the probability method, in original paper, it takes the expectation as the results, however, here we only takes one turn for re-rank evaluation
For the item-side probability evaluation, please ref: https://github.com/XuChen0427/Tax-rank


@inproceedings{xu2024taxation,
  title={A Taxation Perspective for Fair Re-ranking},
  author={Xu, Chen and Ye, Xiaopeng and Wang, Wenjie and Pang, Liang and Xu, Jun and Chua, Tat-Seng},
  booktitle={Proceedings of the 47th International ACM SIGIR Conference on Research and Development in Information Retrieval},
  pages={1494--1503},
  year={2024}
}



"""

class TaxRank(Abstract_Reranker):
    def __init__(self, config, weights = None):
        super().__init__(config, weights)


    def rerank(self, ranking_score, k):
        ## its parameters

        user_size = len(ranking_score)

        t = self.config['t'] #t is from 0 to infty
        assert t>=0
        rerank_list = []

        e = cp.Variable(self.config['item_num'])
        con = [e >= 0, e <= user_size, cp.sum(e) == k * user_size]

        #sorted_ranking_score = np.sort(ranking_score)[::-1]
        eta = k * np.sum(ranking_score, axis=0).reshape(self.config['item_num'])
        item_weight = np.matmul(self.M, self.weights)

        if t == 1:
            fairness = cp.sum(item_weight * cp.multiply(eta, cp.log(e)))
        else:

            fairness = cp.sum(item_weight * cp.multiply(eta, cp.power(e, 1 - t) / (1 - t)))

        obj = cp.Maximize(fairness)

        prob = cp.Problem(obj, con)
        prob.solve()


        e_value = e.value

        group_per_item = np.sum(self.M, axis=0)
        #print(group_per_item)
        item2group_weight = np.matmul(self.M, group_per_item)
        e_value = e_value/item2group_weight
        #print(e_value)
        #exit(0)
        e_value = (e_value/np.sum(e_value)) * k * user_size




        ###start to conduct OT algorithm
        user_size = len(ranking_score)

        Ks = [k for i in range(user_size)]
        answer = ot.sinkhorn(Ks, e_value, 1.0-ranking_score, self.config['ot_lambda'])

        answer = np.where(np.isnan(answer), 0, answer)
        item_prob = np.clip(answer, 0, 1)

        for u in range(user_size):
            mask = np.argsort(ranking_score[u])[:-self.config['sample_num']]
            item_prob[u][mask] = 0.0
            norm_prob = item_prob[u]/np.sum(item_prob[u])

            #print(norm_prob)

            ##  it is the probality method, in original paper, it takes the expectation as the results, however, here we only takes one turn for re-rank
            rerank_items = np.random.choice(self.config['item_num'], size=k, replace=False, p=norm_prob)
            rerank_list.append(rerank_items)

        #recommend_list = answer

        return rerank_list



import numpy as np
import argparse
import yaml
import os


if __name__ == "__main__":
# Initialize ArgumentParser
    parser = argparse.ArgumentParser(description="Fairness in IR systems.")

    # add parameters
    parser.add_argument("--task", type=str, choices=["recommendation", "search"], default='recommendation', help='IR tasks')
    parser.add_argument("--stage", type=str, choices=["pre-processing", "in-processing", "post-processing"],
                        default="in-processing", help="your evaluation stage")
    parser.add_argument("--dataset", type=str, choices=["steam", "clueweb09", "compas"], default="steam", help="your dataset")
    parser.add_argument("--train_config_file", type=str, default="In-processing.yaml", help="your train yaml file")
    #parser.add_argument("--reprocess", type=str, choices=["yes", "no"], default="no", help="your dataset")
    #parser.add_argument("topk", type=float, default=10, help="ranking size")
    args = parser.parse_args()
    with open(os.path.join(args.task, args.train_config_file), 'r') as f:
        train_config = yaml.safe_load(f)
    train_config['dataset'] = args.dataset
    train_config['stage'] = args.stage
    train_config['task'] = args.task
    print("your training config...")
    print(train_config)
    # parse the args

    print("your args:", args)
    if args.task == "recommendation":
        if args.stage == 'in-processing':
            if train_config['use_llm']:
                from recommendation.llm_rec import LLMRecommender
                LLMRecommender = LLMRecommender(train_config)
                LLMRecommender.recommend()
            else:
                from recommendation.trainer import RecTrainer
                trainer = RecTrainer(train_config)
                trainer.train()
        elif args.stage == 'post-processing':
            from recommendation.reranker import RecReRanker
            reranker = RecReRanker(train_config)
            reranker.rerank()
        else:
            raise NotImplementedError("we only support stage in [retrieval, ranking, re-ranking]")
    elif args.task == "search":
        if args.stage == "post-processing":
            if args.dataset != 'clueweb09':
                raise ValueError("For post-processing methods in search, we only support the clueweb09 dataset")
            from search.trainer import SRDTrainer
            trainer = SRDTrainer(train_config)
            trainer.train()
        elif args.stage == "pre-processing":
            from search.trainer_preprocessing_ranker import RankerTrainer
            trainer = RankerTrainer(train_config)
            trainer.train()
    else:
        raise NotImplementedError







import torch
from tqdm import tqdm, trange
import numpy as np
from scipy.sparse import coo_matrix, csr_matrix
from .metric import NDCG, HR, MRR, AUC_score, MMF, Gini, Entropy
import os
import json
from .utils import Build_Adjecent_Matrix
from .metric import *




class Abstract_Evaluator(object):
    def __init__(self, config):
        self.config = config
        self.M, self.iid2pid = Build_Adjecent_Matrix(config)


    def eval(self, dataloader, model, store_scores = False):
        """
            Evaluates the model on the provided dataloader and calculates performance metrics.


            :param dataloader: The data loader that provides batches of user-item interactions and corresponding labels.
            :param model: The model to evaluate.
            :param store_scores: Whether to return the predicted scores as a sparse matrix. Defaults to `False`.
            :return: A dictionary containing the evaluation metric(s) (e.g., AUC score).
        """
        pass


class CTR_Evaluator(Abstract_Evaluator):
    def __int__(self,config):
        super().__init__(config=config)

    def eval(self, dataloader, model, store_scores = False):
        """
            Evaluates the model on the provided dataloader and calculates performance metrics.

            This function runs the evaluation on a dataset using the provided model. It calculates the AUC score based on
            the predicted scores and ground truth labels. If `store_scores` is set to `True`, it also returns the evaluation
            results as a sparse matrix of predicted scores.

            :param dataloader: The data loader that provides batches of user-item interactions and corresponding labels.
            :param model: The model to evaluate.
            :param store_scores: Whether to return the predicted scores as a sparse matrix. Defaults to `False`.
            :return: A dictionary containing the evaluation metric(s) (e.g., AUC score), and optionally, a sparse matrix
                     of predicted scores.
        """
        model.eval()
        y_scores = []
        y_true = []

        row = []
        col = []
        data = []

        with torch.no_grad():
            for user_ids, item_ids, group_ids, label in tqdm(dataloader):
                row.extend(user_ids.numpy().tolist())
                col.extend(item_ids.numpy().tolist())
                user_ids, item_ids = user_ids.to(self.config['device']), item_ids.to(self.config['device'])

                score = model(user_ids, item_ids).cpu().numpy().tolist()
                data.extend(score)
                label = label.cpu().numpy().tolist()
                y_scores.extend(score)
                y_true.extend(label)

        auc_score = AUC_score(y_scores=y_scores, y_true=y_true)
        result_dict = {}
        result_dict["auc"] = np.round(auc_score,self.config['decimals'])

        if store_scores == False:
            return result_dict
        else:
            coo = coo_matrix((data, (row, col)), shape=(self.config['user_num'], self.config['item_num']))
            csr = coo.tocsr() #to remove the zero rows
            csr_eliminated = csr[csr.getnnz(1) > 0]
            coo = csr_eliminated.tocoo()
            return result_dict, coo


class Ranking_Evaluator(Abstract_Evaluator):
    def __int__(self,config):
        super().__init__(config=config)

    def eval(self, dataloader, model, store_scores = False):
        """
            Evaluates the model on the provided dataloader and calculates performance metrics.

            This function runs the evaluation on a dataset using the provided model. It calculates the Ranking metrics based on
            the predicted scores and ground truth labels. If `store_scores` is set to `True`, it also returns the evaluation
            results as a sparse matrix of predicted scores.

            :param dataloader: The data loader that provides batches of user-item interactions and corresponding labels.
            :param model: The model to evaluate.
            :param store_scores: Whether to return the predicted scores as a sparse matrix. Defaults to `False`.
            :return: A dictionary containing the evaluation metric(s), and optionally, a sparse matrix
                     of predicted scores.
        """
        model.eval()
        y_scores = []
        y_true = []

        result_dict = {f"ndcg@{k}":0 for k in self.config['topk']}
        result_dict.update({f"mrr@{k}":0 for k in self.config['topk']})
        result_dict.update({f"hr@{k}":0 for k in self.config['topk']})
        result_dict.update({f"mmf@{k}": 0 for k in self.config['topk']})
        result_dict.update({f"gini@{k}": 0 for k in self.config['topk']})
        result_dict.update({f"entropy@{k}": 0 for k in self.config['topk']})
        result_dict.update({f"maxminratio@{k}": 0 for k in self.config['topk']})
        exposure_dict = {f"top@{k}":np.zeros(self.config['group_num']) for k in self.config['topk']}
        index = 0

        #UI_matrix = np.zeros((self.config['user_num'], self.config['item_num']))
        row = []
        col = []
        data = []

        with torch.no_grad():

            #for user_ids, history_behavior, items, pos_length in tqdm(dataloader):
            for eval_data in tqdm(dataloader):
                user_ids, history_behavior, items, pos_length = eval_data
                batch_size, sample_size = items.shape #item
                #print(items.shape)
                #exit(0)
                pos_length = pos_length.cpu().numpy()

                for b in range(batch_size):
                    row.extend([index]*sample_size)
                    index = index + 1
                    real_item_ids = items[b].numpy().tolist()
                    col.extend(real_item_ids)
                    #print(model.IR_type)
                    if 'retrieval' not in model.IR_type:
                        #if self.config['data_type'] == 'point' or self.config['data_type'] == 'pair':
                        repeat_user_tensor = user_ids[b].repeat(sample_size).unsqueeze(0).to(self.config['device'])
                        #else:
                        repeat_history_tensor = history_behavior[b].repeat(sample_size, 1).unsqueeze(0).to(self.config['device'])

                        user_dict = {"user_ids": repeat_user_tensor,
                                     "history_ids": repeat_history_tensor}
                        i = items[b].to(self.config['device'])
                        score = model(user_dict, i.unsqueeze(0)).cpu().numpy()[0]

                    else:
                        user_dict = {"user_ids":user_ids[b].unsqueeze(0).to(self.config['device']),
                                     "history_ids":history_behavior[b].unsqueeze(0).to(self.config['device'])}
                        i = items[b].to(self.config['device'])

                        score = model.full_predict(user_dict, i.unsqueeze(0)).cpu().numpy()[0]

                    data.extend(score.tolist())
                    #ranked_score = np.sort(score)[::-1]
                    label_list = [1] * pos_length[b] + [0] * (sample_size - pos_length[b])
                    label_list = np.array(label_list)
                    ranked_args = np.argsort(score)[::-1]
                    rank_list = label_list[ranked_args]
                    for k in self.config['topk']:
                        result_dict[f"ndcg@{k}"] += NDCG(rank_list, label_list, k)
                        result_dict[f"mrr@{k}"] += MRR(rank_list, k)
                        result_dict[f"hr@{k}"] += HR(rank_list, label_list, k)

                        ######count the exposures for the computing fairness degree#############
                        ids = ranked_args[:k]
                        rank_items = np.array(real_item_ids)[ids]
                        for i, iid in enumerate(rank_items):
                            group_id = self.iid2pid[iid]
                            if self.config['fairness_type'] == "Exposure":
                                exposure_dict[f"top@{k}"][group_id] += 1
                            else:
                                exposure_dict[f"top@{k}"][group_id] += np.round(score[ids[i]], self.config['decimals'])
                            #exposure_dict[f"top@{k}"][group_id] += 1




        for k in self.config['topk']:
            #print(exposure_dict[f"top@{k}"])
            result_dict[f"mmf@{k}"] = MMF(exposure_dict[f"top@{k}"], ratio=self.config['mmf_eval_ratio']) * index
            result_dict[f"gini@{k}"] = Gini(exposure_dict[f"top@{k}"]) * index
            result_dict[f"entropy@{k}"] = Entropy(exposure_dict[f"top@{k}"]) * index
            result_dict[f"maxminratio@{k}"] = MinMaxRatio(exposure_dict[f"top@{k}"]) * index


        for key in result_dict.keys():
            result_dict[key] = np.round(result_dict[key]/index, self.config['decimals'])

        if store_scores == False:
            return result_dict
        else:
            return result_dict, coo_matrix((data, (row, col)), shape=(index, self.config['item_num']))



class LLM_Evaluator(Abstract_Evaluator):
    def __init__(self, config):
        super().__init__(config=config)
        self.topk_list = config['topk']

    def get_data(self, data):
        """
        This method processes the input data to extract prediction lists, label lists, and score lists for each user.

        :param 'predict_list': A list of predicted items.
        :param 'positive_items': A list of items that are considered positive (e.g., liked or preferred by the user).
        :param 'scores': A list of scores corresponding to the predicted items, indicating the confidence of the prediction.

        :return:
        - `predict_lists`: A list of predict lists for all users.
        - `label_lists`: For each user, a list of binary labels indicating whether each predicted item is positive (1) or not (0).
        - `score_lists`: A list of score lists corresponding to the predicted items for all users.
        """
        # ground_truths = [i['positive_items'] for i in data]
        # sens_feat = [i['sensitiveAttribute'] for i in data]
        label_lists = []
        # ranking_lists = []
        score_lists = []
        predict_lists = []
        for user in data:
            p = user['predict_list']
            predict_lists.append(p)
            label_list = [1 if m in user['positive_items'] else 0 for m in p]
            # label_list = [1 if m in user['positive_items'] else 0 for m in user['item_candidates']]
            score = user['scores']
            score_lists.append(score)
            label_lists.append(label_list)
            # ranking_lists.append(ranking_list)

        return predict_lists, label_lists, score_lists

    def get_cates_value(self, iid2pid, predict, topk):
        """
        Get the category values based on predicted indices and their corresponding categories.

        This method processes the predicted indices along with their mapping to category IDs
        and returns a list of counts for each category, representing the frequency of occurrence
        in the top-k predictions.

        :param iid2pid : dict
            A dictionary mapping item indices (int) to their respective category IDs (int).
            If an item index is not found in the dictionary, it defaults to -1.

        :param predict : List[List[int]]
            A 2D list where each sublist contains the predicted indices (top-k predictions)
            for corresponding input data points.

        :param topk : int
            The number of top predictions considered for each data point. This determines
            how many elements from the beginning of each sublist in `predict` are processed.

        :return: List[int]
            A list of integers where each value corresponds to the count of occurrences for
            a specific category across all top-k predictions. The order of these counts matches
            the sorted order of category IDs as returned by `get_categories(iid2pid)`.
        """
        cates_name = self.get_categories(iid2pid)
        predict = [i[:topk] for i in predict]
        from collections import defaultdict
        cates_count = defaultdict(int)
        for p in predict:
            for prediction in p:
                c = iid2pid.get(prediction, -1)
                cates_count[c] += 1  # not score-based scores[idx][k]
        values = [cates_count[i] for i in cates_name]
        return values

    def cal_acc_score(self, label_lists, score_lists, topk):
        """
        Calculate accuracy scores for recommendation system evaluation.

        This method computes the average NDCG (Normalized Discounted Cumulative Gain), HR (Hit Ratio), and MRR (Mean Reciprocal Rank)
        at a specified `topk` cutoff for a list of ground-truth labels and corresponding prediction scores.

        :param label_lists : List[List[int]]
            A list of lists containing ground-truth labels. Each sublist represents the relevant items for a user or query.
        :param score_lists : List[List[float]]
            A list of lists containing predicted scores. Each sublist corresponds to the relevance scores for items
            matching the order in `label_lists`.
        :param topk : int
            The number of top predictions to consider when calculating the metrics.

        :return: Dict[str, float]
            A dictionary containing the average NDCG, HR, and MRR scores at the given `topk`, with keys formatted as 'NDCG@{topk}',
            'HR@{topk}', and 'MRR@{topk}' respectively. Scores are rounded to 4 decimal places.
        """
        score = {}
        ndcgs = []
        hrs = []
        mrrs = []
        for lab, sco in zip(label_lists, score_lists):
            ndcg = NDCG(lab, lab, topk)
            hr = HR(lab, lab, topk)
            mrr = MRR(lab, topk)
            ndcgs.append(ndcg)
            hrs.append(hr)
            mrrs.append(mrr)

        # compute metrics
        score[f'NDCG@{topk}'] = np.round(np.mean(ndcgs), 4)
        score[f'HR@{topk}'] = np.round(np.mean(hrs), 4)
        score[f'MRR@{topk}'] = np.round(np.mean(mrrs), 4)
        return score

    def get_categories(self, iid2pid):
        return list(set(iid2pid.values()))

    def cal_fair_score(self, iid2pid, predict, topk):
        """
        Calculate fairness scores for a recommendation system's evaluation.

        This method computes various fairness metrics at a specified top-k cutoff to evaluate
        the diversity and inclusiveness of the predicted items. It utilizes different fairness
        measures like MMF (Max-Min Fairness), Gini coefficient, Min-Max Ratio, and Entropy to
        quantify the balance across different categories or groups within the predictions.

        :param iid2pid (Dict[int, int]): A mapping where keys are item IDs and values are their respective group/category IDs.
        :param predict (List[Tuple[int, float]]): A list of tuples, each containing an item ID and its predicted score/score.
        :param topk (int): The top-k count used to consider the highest scored items for fairness evaluation.

        :return:
        A dictionary with keys as the metric names prefixed with the top-k cutoff (e.g., 'MMF@5') and values as the
        corresponding calculated scores, rounded to 4 decimal places.
        """
        #
        score = {}
        cates_value = self.get_cates_value(iid2pid, predict, topk)
        # print(cates_value)
        mmf = MMF(cates_value)
        cate_gini = Gini(cates_value)
        maxmin_ratio = MinMaxRatio(cates_value)
        # cv = (cates_value)
        entropy = Entropy(cates_value)
        score[f'MMF@{topk}'] = np.round(mmf, 4)
        score[f'Gini@{topk}'] = np.round(cate_gini, 4)
        score[f'MMR@{topk}'] = np.round(maxmin_ratio, 4)
        # score[f'cv@{topk}'] = np.round(cv, 4)
        score[f'Entropy@{topk}'] = np.round(entropy, 4)
        return score

    def llm_eval(self, grounding_result, iid2pid):
        """
        Evaluate the performance of a language model based on grounding results and item-pid mappings.

        This method assesses the accuracy and fairness of the model's predictions at different top-K thresholds.
        It computes both accuracy scores and fairness scores,汇总 these into a comprehensive evaluation result.

        :param grounding_result (Dict[str, Any]): The output from the model grounding process, containing necessary information for evaluation.
        :param iid2pid (Dict[str, str]): A mapping from item IDs to product IDs, used in calculating fairness metrics.

        :return:
        - eval_result (Dict[str, float]): A dictionary summarizing the evaluation outcomes, including accuracy and fairness scores for each specified top-K value.
        """
        predict_lists, label_lists, score_lists = self.get_data(grounding_result)
        eval_result = {}
        for topk in self.topk_list:
            acc_score = self.cal_acc_score(label_lists, score_lists, topk)
            fair_score = self.cal_fair_score(iid2pid, predict_lists, topk)
            eval_result.update(acc_score)
            eval_result.update(fair_score)
            # eval_result.update({f'Top{topk}': acc_score})
        print(f'Evaluate_result:{eval_result}')
        return eval_result


"""
@article{yang2020causal,
  title={Causal intersectionality for fair ranking},
  author={Yang, Ke and Loftus, Joshua R and Stoyanovich, Julia},
  journal={arXiv preprint arXiv:2006.08688},
  year={2020}
}

License
Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import os
from .modules.CIFRank.run_causal_model import run_causal_model
from .modules.CIFRank.generate_counterfactual_data import get_counterfactual_data_real
from .fair_model import PreprocessingFairnessIntervention

class CIFRank(PreprocessingFairnessIntervention):
    """
        CIFRank class implements a fairness intervention method using causal intersectionality.

        This class extends the PreprocessingFairnessIntervention class and applies a causal fairness
        ranking model to the given dataset.
    """
    def __init__(self, configs, dataset):
        """
                Initialize the CIFRank model with configuration settings and dataset information.

                :param configs : dict
                    Configuration dictionary containing model parameters.
                :param dataset : str
                    The name or path of the dataset to be processed.
        """
        super().__init__(configs, dataset)

    def fit(self, X_train, run):
        """
        Train the causal fairness ranking model on the given training dataset.

        This method ensures that the necessary model directory exists and then runs the causal model.
        The trained model is saved in `self.model_path`.

        :param X_train : pandas.DataFrame
            The training dataset.
        :param run : str
            The identifier for the training run.

        :return : None
        """
        if not os.path.exists(os.path.join(self.model_path, run)):
            os.makedirs(os.path.join(self.model_path, run))
            run_causal_model(X_train, self.configs, self.dataset, os.path.join(self.model_path, run))

    def transform(self, X, run, file_name=None):
        """
        Apply the fairness transformation to the dataset by generating fair counterfactual columns.

        Fair columns are generated using a naming convention `<column_name>_fair` and saved to a CSV file
        if it does not already exist.

        :param X : pandas.DataFrame
            The dataset to which the fairness method should be applied.
        :param run : str
            The identifier for the transformation run.
        :param file_name : str, optional
            Name of the file to save the transformed dataset.

        :return : pandas.DataFrame
            The dataset with appended fair columns.
        """
        file_path = os.path.join(self.fair_data_path, run, f'fair_{file_name}_data.csv')
        if not os.path.exists(file_path):
            data_fair = self.generate_counterfactual_data(X, run)
            if file_name !=None:
                os.makedirs(os.path.join(self.fair_data_path, run), exist_ok=True)
                data_fair.to_csv(file_path)

        return data_fair


    def generate_counterfactual_data(self, X, run):
        """Generates the fair data by using the causal estimates
        Args:
            data (pandas.Dataframe): data on which to append the fair columns

        Returns:
            data (pandas.Dataframe): data containing the appended fair columns which contain
            the counterfactual values computed based on the causal estimates
        """
        path_causal = os.path.join(self.model_path, run)
        if os.path.exists(path_causal):
            if len(os.listdir(path_causal)):
                fair_df = get_counterfactual_data_real(X, path_causal, self.configs, self.dataset)

        return fair_df


import os
import yaml


class SRDTrainer(object):
    def __init__(self, train_config):
        """
        Initialize post-processing and base models.
        :param train_config: Your custom config files.
        """

        self.train_config = train_config

    def load_configs(self, dir):
        """
        Loads and merges configuration files for the model, dataset, and evaluation.

        This function loads multiple YAML configuration files, including the process configuration,
        dataset-specific settings, model configurations, and evaluation parameters. All configurations
        are merged, with the highest priority given to the class's own `config` attribute.

        :param dir: The directory where the main process configuration file is located.
        :return: A dictionary containing the merged configuration from all files.
        """

        print("start to load dataset config...")
        with open(os.path.join(self.train_config['task'], "properties", "dataset", self.train_config['dataset'].lower() + ".yaml"), 'r') as f:
            config = yaml.safe_load(f)
        config.update({'data_dir': dir})

        print("start to load model config...")

        with open(os.path.join(self.train_config['task'], "properties", "models", self.train_config['model'].lower() + ".yaml"),
                  'r') as f:
            model_config = yaml.safe_load(f)

        config.update(model_config)

        with open(os.path.join("search", "properties", "evaluation.yaml"), 'r') as f:
            config.update(yaml.safe_load(f))

        config.update(self.train_config)  ###train_config has highest rights
        print("your loading config is:")
        print(config)

        return config


    def train(self):
        """
        Training post-processing search model main workflow.
        """

        dir = os.path.join(self.train_config['task'], "processed_dataset", self.train_config['dataset'])
        config = self.load_configs(dir)

        if os.path.exists(os.path.join(config['task'], "processed_dataset", config['dataset'], config['model'])) and config['reprocess'] == False:
            print("Data has been processed, start to load the dataset...")
        else:
            print("start to process data...")
            if os.path.join(config['data_dir'], 'div_query.data') not in config['data_dir']:
                from .utils.process_dataset import data_process
                data_process(config)
            if config['model'].lower() == 'desa':
                from .utils.process_desa import Process
                Process(config)
            elif config['model'].lower() == 'daletor':
                from .utils.process_daletor import Process
                Process(config)
            elif config['model'].lower() == 'xquad':
                from .utils.process_bm25 import generate_bm25_scores_for_query
                generate_bm25_scores_for_query(config)
            elif config['model'].lower() == 'pm2':
                from .utils.process_bm25 import generate_bm25_scores_for_query
                generate_bm25_scores_for_query(config)
            elif config['model'].lower() == 'llm':
                pass
            else:
                raise NotImplementedError(f"Not supported model type: {config['model']}")

        print("start to load dataset......")
        self.device = config['device']
        if config['mode'] == 'test' and config['best_model_list'] != []:
            print("start to test the model...")
            from .post_evaluator import get_global_fullset_metric
            get_global_fullset_metric(config)
        elif config['mode'] == 'train':
            """
            For implementing your own supervised methods, you need to first re-write xxx_run and then implement your own model.
            You can also just copy the two example xxx_run implementation for quick starting.
            For the unsupervised methods, just implment the model function.
            """

            if config['model'].lower() == 'desa':
                from .datasets.DESA import DESA_run
                DESA_run(config)
            elif config['model'].lower() == 'daletor':
                from .datasets.DALETOR import DALETOR_run
                DALETOR_run(config)
            elif config['model'].lower() == 'xquad':
                from .postprocessing_model.xQuAD import xQuAD
                xquad = xQuAD()
                xquad.rerank(config)
            elif config['model'].lower() == 'pm2':
                from .postprocessing_model.PM2 import PM2
                pm2 = PM2()
                pm2.rerank(config)
            elif config['model'].lower() == 'llm':
                from .datasets.LLM import llm_run
                llm_run(config)
            else:
                raise NotImplementedError(f"Not supported model type: {config['model']}")






import os
import copy
import pickle
import multiprocessing
import xml.dom.minidom

from .div_type import div_query
from .utils import split_list

MAXDOC = 50
REL_LEN = 18


def get_query_dict(config):
    """
    Generates a dictionary of queries and their subtopics.
    :param config: A dictionary containing configuration settings.
    :return: A dictionary mapping query IDs (qid) to `div_query` objects containing the query and subtopics.
    """
    dq_dict = {}
    topics_list = []
    for year in ['2009','2010','2011','2012']:
        filename = os.path.join(config['data_dir'], config['ground_truth'], 'wt_topics/wt' + year + '.topics.xml')
        DOMTree = xml.dom.minidom.parse(filename)
        collection = DOMTree.documentElement
        topics = collection.getElementsByTagName("topic")
        topics_list.extend(topics)
    ''' load subtopics for each query '''
    for topic in topics_list:
        if topic.hasAttribute("number"):
            qid = topic.getAttribute("number")
        query = topic.getElementsByTagName('query')[0].childNodes[0].data
        subtopics = topic.getElementsByTagName('subtopic')
        subtopic_id_list = []
        subtopic_list = []
        for subtopic in subtopics:
            if subtopic.hasAttribute('number'):
                subtopic_id = subtopic.getAttribute('number')
                subtopic_id_list.append(subtopic_id)
            sub_query = subtopic.childNodes[0].data
            subtopic_list.append(sub_query)
        dq = div_query(qid, query, subtopic_id_list, subtopic_list)
        dq_dict[str(qid)] = dq
    return dq_dict


def get_query_suggestion(dq, config):
    """
    Adds query suggestions to the query dictionary (dq) for each query.
    
    :param dq: A dictionary of `div_query` objects.
    :param config: A dictionary containing configuration settings.
    :return: A dictionary of `div_query` objects with added query suggestions.
    """
    dq_dict = {}
    filename = os.path.join(config['data_dir'], 'query_suggestion.xml')
    DOMTree = xml.dom.minidom.parse(filename)
    collection = DOMTree.documentElement
    topics = collection.getElementsByTagName("topic")
    ''' load subtopics for each query '''
    for topic in topics:
        if topic.hasAttribute("number"):
            qid = topic.getAttribute("number")
        query = topic.getElementsByTagName('query')[0].childNodes[0].data
        subtopics = topic.getElementsByTagName('subtopic1')
        subtopic_id_list = []
        subtopic_list = []
        for subtopic in subtopics:
            if subtopic.hasAttribute('number'):
                subtopic_id = subtopic.getAttribute('number')
                subtopic_id_list.append(subtopic_id)
            suggestion = subtopic.getElementsByTagName('suggestion')[0].childNodes[0].data
            subtopic_list.append(suggestion)
        dq[str(qid)].add_query_suggestion(subtopic_list)
    return dq_dict


def get_docs_dict(config):
    """ Loads the document IDs and their relevance scores for each query.
    docs_dict[qid] = [doc_id, ...]
    docs_rel_score_dict[qid] = [score, ...]

    :param config: A dictionary containing configuration settings.
    :return: Two dictionaries: docs_dict (query ID to document IDs) and docs_rel_score_dict (query ID to relevance scores).
    """
    docs_dict = {}
    docs_rel_score_dict = {}
    for year in ['2009','2010','2011','2012']:
        filename = os.path.join(config['data_dir'], config['ground_truth'], 'wt' + year + '.txt')
        f = open(filename)
        for line in f:
            qid, _, docid, _, score, _ = line.split(' ')
            if str(qid) not in docs_dict:
                docs_dict[str(qid)] = []
                docs_rel_score_dict[str(qid)] = []
            docs_dict[str(qid)].append(str(docid))
            docs_rel_score_dict[str(qid)].append(float(score))
    ''' Normalize the relevance score of the documents '''
    for qid in docs_rel_score_dict:
        temp_score_list = copy.deepcopy(docs_rel_score_dict[qid])
        for i in range(len(temp_score_list)):
            temp_score_list[i] = docs_rel_score_dict[qid][0]/docs_rel_score_dict[qid][i]
        docs_rel_score_dict[qid] = temp_score_list
    return docs_dict, docs_rel_score_dict


def get_doc_judge(qd, dd, ds, config):
    """ Loads the document lists and relevance score lists for the corresponding queries.
    
    :param qd: A dictionary of `div_query` objects.
    :param dd: A dictionary of document IDs for each query.
    :param ds: A dictionary of relevance scores for documents for each query.
    :param config: A dictionary containing configuration settings.
    :return: The updated `qd` dictionary with documents and relevance scores added, and judged for relevance.
    """
    get_query_suggestion(qd, config)
    for key in qd:
        qd[key].add_docs(dd[key])
        qd[key].add_docs_rel_score(ds[key])
    for year in ['2009','2010','2011','2012']:
        filename = os.path.join(config['data_dir'], config['ground_truth'], 'wt_judge/' + year + '.diversity.qrels')
        f = open(filename, 'r')
        for line in f:
            qid, subtopic, docid, judge = line.split(' ')
            judge = int(judge)
            if judge > 0:
                if str(docid) in qd[str(qid)].subtopic_df.index.values:
                    qd[str(qid)].subtopic_df[str(subtopic)][str(docid)] = 1
    return qd


def data_process_worker(task, data_dir):
    """
    Processes a list of queries and saves them as data files.
    
    :param task: A list of query ID and `div_query` objects to process.
    :param data_dir: The directory where processed query data will be saved.
    """
    for item in task:
        qid = item[0]
        dq = item[1]
        ''' get the best ranking for the top 50 relevant documents '''
        dq.get_best_rank(MAXDOC)
        pickle.dump(dq, open(data_dir+str(qid)+'.data', 'wb'), True)


def calculate_best_rank(qd, config):
    """
    Calculates the best ranking of documents for each query.
    
    :param qd: A dictionary of `div_query` objects.
    :param config: A dictionary containing configuration settings.
    """
    data_dir = os.path.join(config['data_dir'], 'best_rank/')
    if not os.path.exists(data_dir):
        os.makedirs(data_dir)
    q_list = []
    for key in qd:
        x = copy.deepcopy(qd[key])
        q_list.append((str(key), x))
    jobs = []
    task_list = split_list(q_list, 8)
    for task in task_list:
        p = multiprocessing.Process(target = data_process_worker, args = (task, data_dir))
        jobs.append(p)
        p.start()
    for job in jobs:
        job.join()


def get_stand_best_metric(qd, config):
    """
    Loads the best alpha-nDCG metric from the DSSA.
    
    :param qd: A dictionary of `div_query` objects.
    :param config: A dictionary containing configuration settings.
    """
    std_dict = pickle.load(open(os.path.join(config['data_dir'], 'stand_metrics.data'), 'rb'))
    for qid in std_dict:
        m = std_dict[qid]
        target_q = qd[str(qid)]
        target_q.set_std_metric(m)


def generate_qd(config):
    """
    Generates a `div_query` file from the data directory.
    
    :param config: A dictionary containing configuration settings.
    :return: A dictionary of `div_query` objects.
    """
    data_dir = os.path.join(config['data_dir'], 'best_rank/')
    files = os.listdir(data_dir)
    files.sort(key = lambda x:int(x[:-5]))
    query_dict = {}
    for f in files:
        file_path = os.path.join(data_dir, f)
        temp_q = pickle.load(open(file_path, 'rb'))
        query_dict[str(f[:-5])] = temp_q
    pickle.dump(query_dict, open(os.path.join(config['data_dir'], 'div_query.data'), 'wb'), True)
    return query_dict
    

def data_process(config):
    """ Main function for processing query, document, and relevance data.
    :param config: A dictionary containing configuration settings.
    """
    ''' get subtopics for each query '''
    qd = get_query_dict(config)
    ''' get documents dictionary '''
    dd, ds = get_docs_dict(config)
    ''' get diversity judge for documents '''
    qd = get_doc_judge(qd, dd, ds, config)
    ''' get the stand best alpha-nDCG from DSSA '''
    get_stand_best_metric(qd, config)
    ''' get the best ranking for top n relevant documents and save as files'''
    calculate_best_rank(qd, config)
    ''' combine the best ranking into a single file '''
    generate_qd(config)



import torch

MAXDOC = 50


def list_pairwise_loss(score_1, score_2, delta):
    """
    Computes the accuracy and list-pairwise loss based on the difference between two scores. 
    
    :param score_1: A tensor containing the scores for the first set of items.
    :param score_2: A tensor containing the scores for the second set of items.
    :param delta: A tensor representing the weight for each pairwise comparison.
    :return: The accuracy and the computed list-pairwise loss.
    """
    acc = torch.Tensor.sum((score_1-score_2)>0).item()/float(score_1.shape[0])
    loss = -torch.sum(delta * torch.Tensor.log(1e-8+torch.sigmoid(score_1 - score_2)))/float(score_1.shape[0])
    return acc, loss


def ndcg_loss(score, div):
    """
    Computes the NDCG (Normalized Discounted Cumulative Gain) loss for ranking tasks. 
    
    :param score: A tensor containing the predicted scores for each document.
    :param div: A tensor containing the relevance or diversity values for each document.
    :return: The computed NDCG loss.
    """
    T = 0.1
    temp_list = [(score[i,:] - score[i,:].reshape(1,MAXDOC).permute(1,0))/T for i in range(score.shape[0])]
    temp = torch.stack(temp_list, dim=0).float()
    sigmoid_temp = torch.sigmoid(temp)
    C_li = torch.bmm(sigmoid_temp, div) - 0.5 * div
    pw = torch.pow(0.5, C_li)
    top = torch.mul(div, pw)
    top = torch.sum(top, dim=-1)
    R_i = 0.5 + torch.sum(sigmoid_temp, dim = -1)
    bottom = torch.log2(1+R_i)
    #loss = 1.0 / (1e-6 + torch.sum(top/bottom))
    loss = - torch.sum(top/bottom)
    print('loss = ', loss)
    return loss






"""
Implementation of the ICDE 2019 paper
iFair_module: Learning Individually Fair Data Representations for Algorithmic Decision Making
url: https://ieeexplore.ieee.org/document/8731591
citation:
@inproceedings{DBLP:conf/icde/LahotiGW19,
  author    = {Preethi Lahoti and
               Krishna P. Gummadi and
               Gerhard Weikum},
  title     = {iFair_module: Learning Individually Fair Data Representations for Algorithmic
               Decision Making},
  booktitle = {35th {IEEE} International Conference on Data Engineering, {ICDE} 2019,
               Macao, China, April 8-11, 2019},
  pages     = {1334--1345},
  publisher = {{IEEE}},
  year      = {2019},
  url       = {https://doi.org/10.1109/ICDE.2019.00121},
  doi       = {10.1109/ICDE.2019.00121},
  timestamp = {Wed, 16 Oct 2019 14:14:56 +0200},
  biburl    = {https://dblp.org/rec/conf/icde/LahotiGW19.bib},
  bibsource = {dblp computer science bibliography, https://dblp.org}
}


__author__: Preethi Lahoti
__email__: plahoti@mpi-inf.mpg.de
"""

import os
import numpy as np
import pandas as pd
from scipy.optimize import minimize
from .modules.iFair.loss import iFair_optimisation as ifair_func
from .modules.probabilistic_mapping_helpers import compute_X_hat, compute_euclidean_distances
from .utils import process_data_input, process_data_output, save_model_data,load_model_data
from .fair_model import PreprocessingFairnessIntervention


class iFair(PreprocessingFairnessIntervention):
    """
        iFair is a fairness intervention method based on optimization techniques.

        This class extends PreprocessingFairnessIntervention and applies individual fairness constraints to data
        using probabilistic mapping and distance-based optimization.
    """
    def __init__(self, configs, dataset):
        """
            Initialize the iFair model with the given configuration settings and dataset.

            :param configs : dict
                Configuration dictionary containing model parameters.
            :param dataset : str
                The name or path of the dataset to be processed.
        """
        super().__init__(configs, dataset)

    def fit(self, X_train, run):
        """
        Train the iFair fairness model using the given training dataset.

        This method applies optimization to learn individual fairness constraints and stores the results
        for later use.

        :param X_train : pandas.DataFrame
            The training dataset. The last column is expected to be the protected attribute.
        :param run : str
            The identifier for the training run.

        :return : None
        """
        if not os.path.exists(os.path.join(self.model_path, run)):
            X_train, group_weights, sensitive_groups, sensitive_column_indices, nonsensitive_column_indices = (
                process_data_input(X_train, self.configs, self.dataset))

            if self.configs["seed"] is not None:
                np.random.seed(self.configs["seed"])


            D_X_F = compute_euclidean_distances(X_train, nonsensitive_column_indices)
            l = len(nonsensitive_column_indices)

            P = X_train.shape[1]
            min_obj = None
            opt_params = None
            for i in range(self.configs["nb_restarts"]):
                x0_init = np.random.uniform(size=int(P + P * self.configs["k"]))
                # setting protected column weights to epsilon
                ## assumes that the column indices from l through P are protected and appear at the end
                for i in range(l, P, 1):
                    x0_init[i] = 0.0001
                bnd = [(0, 1)] * P + [(None, None)] * P * self.configs["k"]

                ifair_func.iters = 0
                opt_result = minimize(ifair_func, x0_init,
                                      args=(X_train, D_X_F, self.configs["k"],
                                            self.configs["A_x"], self.configs["A_z"], os.path.join(self.model_path, run)),
                                      method='L-BFGS-B',
                                      jac=False,
                                      bounds=bnd,
                                      options={'maxiter': self.configs["max_iter"],
                                               'maxfun': self.configs["maxfun"],
                                               'eps': 1e-3})

            if (min_obj is None) or (opt_result.fun < min_obj):
                min_obj = opt_result.fun
                opt_params = opt_result.x

            self.opt_params = opt_params
            save_model_data(self, os.path.join(self.model_path, run))
        else:
            self.opt_params = load_model_data(os.path.join(self.model_path, run))

    def transform(self, X, run, file_name=None):
        """
            Apply the fairness transformation to the dataset using the learned model.

            This method ensures fairness by adjusting feature distributions while maintaining data utility.

            :param X : pandas.DataFrame
                The dataset to which the fairness transformation is applied.
            :param run : str
                The identifier for the transformation run.
            :param file_name : str, optional
                Name of the file to save the transformed dataset.

            :return : pandas.DataFrame
                The dataset with transformed fair columns.
        """
        fair_data_path = os.path.join(self.fair_data_path, run)
        os.makedirs(fair_data_path, exist_ok=True)
        fair_data_file = os.path.join(fair_data_path, f'fair_{file_name}_data.csv')
        if not os.path.exists(fair_data_file):
            X_np, group_weights, sensitive_groups, sensitive_column_indices, nonsensitive_column_indices = (
                process_data_input(X, self.configs, self.dataset))

            X_hat, _ = compute_X_hat(X_np, self.opt_params, self.configs["k"], alpha=True)
            X_fair = process_data_output(X, X_hat, self.dataset, nonsensitive_column_indices, fair_data_path, file_name)
        else:
            X_fair = pd.read_csv(fair_data_file)
        return X_fair







import numpy as np
from .Abstract_Ranker import Abstract_Sampler
import torch.nn.functional as F
import torch
import random

r"""
FairNeg
################################################
@inproceedings{10.1145/3543507.3583355,
author = {Chen, Xiao and Fan, Wenqi and Chen, Jingfan and Liu, Haochen and Liu, Zitao and Zhang, Zhaoxiang and Li, Qing},
title = {Fairly Adaptive Negative Sampling for Recommendations},
year = {2023},
isbn = {9781450394161},
publisher = {Association for Computing Machinery},
address = {New York, NY, USA},
url = {https://doi.org/10.1145/3543507.3583355},
doi = {10.1145/3543507.3583355},
abstract = {Pairwise learning strategies are prevalent for optimizing recommendation models on implicit feedback data, which usually learns user preference by discriminating between positive (i.e., clicked by a user) and negative items (i.e., obtained by negative sampling). However, the size of different item groups (specified by item attribute) is usually unevenly distributed. We empirically find that the commonly used uniform negative sampling strategy for pairwise algorithms (e.g., BPR) can inherit such data bias and oversample the majority item group as negative instances, severely countering group fairness on the item side. In this paper, we propose a Fairly adaptive Negative sampling approach (FairNeg), which improves item group fairness via adaptively adjusting the group-level negative sampling distribution in the training process. In particular, it first perceives the model’s unfairness status at each step and then adjusts the group-wise sampling distribution with an adaptive momentum update strategy for better facilitating fairness optimization. Moreover, a negative sampling distribution Mixup mechanism is proposed, which gracefully incorporates existing importance-aware sampling techniques intended for mining informative negative samples, thus allowing for achieving multiple optimization purposes. Extensive experiments on four public datasets show our proposed method’s superiority in group fairness enhancement and fairness-utility tradeoff.},
booktitle = {Proceedings of the ACM Web Conference 2023},
pages = {3723–3733},
numpages = {11},
keywords = {BPR, Fairness, Negative Sampling., Recommender Systems},
location = {Austin, TX, USA},
series = {WWW '23}
}
"""


class FairNeg(Abstract_Sampler):
    def __init__(self, config, user2pos):
        super().__init__(config, user2pos)

        self.sample_probality = np.ones(self.config['group_num'])
        self.sample_probality = self.sample_probality / np.sum(self.sample_probality)

        self.history_gradient = np.zeros(self.config['group_num'])

        self.epoch_loss = np.zeros(self.config['group_num'])

    def reset_parameters(self):

        gradient = self.epoch_loss -  np.mean(self.epoch_loss)
        if np.sum(gradient) == 0:
            pass
        else:
            gradient = gradient / np.sum(gradient)
            v = self.config['gamma'] * self.history_gradient + self.config['alpha'] * gradient
            v = v/ np.sum(v)
            self.history_gradient = v
            self.sample_probality = self.sample_probality - v
            self.sample_probality = self.sample_probality / (np.sum(self.sample_probality)+1e-5)
            self.epoch_loss = np.zeros(self.config['group_num'])


    def accumulate_epoch_loss(self, delta_epoch_loss):
        self.epoch_loss += delta_epoch_loss


    def sample(self, interaction, Model):
        user_ids = interaction['user_ids'].cpu().numpy().tolist()
        item_ids = interaction['item_ids'].cpu().numpy()
        neg_items = []

        pos_items = []
        for u in user_ids:
            pos = self.user2pos[u]
            pos_items.extend(pos)
        pos_items = set(pos_items)
        neg_item_corpus = set(range(self.config['item_num'])) - pos_items
        neg_item_corpus = random.sample(list(neg_item_corpus), self.config['neg_sample_num'])


        group_fair = self.sample_probality / (self.group_sizes + 1)
        prob_fair = np.matmul(self.M[neg_item_corpus], group_fair)

        scores = Model.full_scores(interaction, torch.tensor(neg_item_corpus,dtype=torch.long).to(interaction['user_ids'].device))
        scores = torch.softmax(scores/self.config['temp'], dim=-1)
        scores = scores.detach().cpu().numpy()
        for i, user in enumerate(user_ids):

            prob = self.config['beta'] * prob_fair + (1-self.config['beta']) * scores[i]
            prob = prob/np.sum(prob+1e-5)
            #print(prob)
            neg_item = random.choices(neg_item_corpus, weights=prob, k=1)[0]
            neg_items.append(neg_item)

        return torch.tensor(neg_items, dtype=torch.long).to(interaction['user_ids'].device), self.M[item_ids]











import os
import numpy as np
from .Abstract_Reranker import Abstract_Reranker
from tqdm import tqdm,trange

r"""
CPFair

Note that we extend it into multi-group cases and only focus on item-side, slightly different from the original paper

@inproceedings{naghiaei2022cpfair,
  title={Cpfair: Personalized consumer and producer fairness re-ranking for recommender systems},
  author={Naghiaei, Mohammadmehdi and Rahmani, Hossein A and Deldjoo, Yashar},
  booktitle={Proceedings of the 45th International ACM SIGIR Conference on Research and Development in Information Retrieval},
  pages={770--779},
  year={2022}
}
"""

class CPFair(Abstract_Reranker):
    def __init__(self, config, weights = None):
        super().__init__(config, weights)


    def rerank(self, ranking_score, k):
        user_size = len(ranking_score)
        B_l = np.zeros(self.group_num)
        lambd = self.config['lambda']
        rerank_list = []

        for u in trange(user_size):
            minimax_reg = lambd * np.matmul(self.M,1-(B_l/(np.sum(B_l)+1e-5)))
            rel = ranking_score[u,:] + minimax_reg
            result_item = np.argsort(rel)[::-1]
            result_item = result_item[:k]
            rerank_list.append(result_item)
            B_l = B_l + np.sum(self.M[result_item,:],axis=0,keepdims=False)


        return rerank_list



"""
Implementation of the ICDE 2019 paper
iFair_module: Learning Individually Fair Data Representations for Algorithmic Decision Making
__url__: https://ieeexplore.ieee.org/abstract/document/8731591
__author__: Preethi Lahoti
__email__: plahoti@mpi-inf.mpg.de
"""

import numpy as np
import os
from scipy import linalg
from sklearn.metrics import pairwise
from ..probabilistic_mapping_helpers import compute_X_hat
def compute_reconstruction_loss(x_orig, x_hat):
    """Computes the reconstruction loss (L_x)."""
    return np.mean((x_hat - x_orig) ** 2)

def compute_individual_fairness_loss(D_X_f, X_hat):
    """Computes the individual fairness loss (L_z)."""
    D_X_f_hat = pairwise.euclidean_distances(X_hat, X_hat)
    L_z = linalg.norm(D_X_f - D_X_f_hat)
    return L_z


def iFair_optimisation(params, X,D_X_f=None, k=10, A_x=1e-4, A_z=1e-4, logs_path=""):
    """iFair optimisation function."""
    iFair_optimisation.iters += 1
    X_hat, _ = compute_X_hat(X, params, k, alpha=True)

    L_x = compute_reconstruction_loss(X, X_hat)
    L_z = compute_individual_fairness_loss(D_X_f, X_hat)

    criterion = A_x * L_x + A_z * L_z

    if iFair_optimisation.iters % 100 == 0:
        print("step: {}, L_x: {},  L_z: {}, loss:{}\n".format(
                    iFair_optimisation.iters, L_x, L_z, criterion))
        if logs_path != '':
            os.makedirs(logs_path, exist_ok=True)
            with open(os.path.join(logs_path, "logs.txt"), 'a') as f:
                f.write("step: {}, L_x: {},  L_z: {}, loss:{}\n".format(
                    iFair_optimisation.iters, L_x, L_z, criterion))
    return criterion



import numpy as np
import pandas as pd
import math
import os
from sklearn.metrics import pairwise


def evaluate_runs(args_eval, qids, eval_path, runs, split):
    """
    Evaluate the runs by calculating metrics and saving the results in CSV files.

    :param args_eval : dict
        A dictionary containing evaluation configurations, including metrics and other settings.
    :param qids : list
        A list of unique query identifiers (QIDs) to evaluate.
    :param eval_path : str
        Path where evaluation results will be saved.
    :param runs : int
        Number of runs to evaluate.
    :param split : str
        The data split to evaluate (e.g., "train", "test").

    :return : None
        This function saves the evaluation results as CSV files.
    """
    for qid in qids:
        for eval_measure in args_eval['metrics']:
            output_f = os.path.join(eval_path, "Eval_QID_" + str(qid) + '_' + eval_measure + ".csv")

            if not os.path.exists(output_f):
                res_all = []
                for r in range(runs):
                    file_name = 'Eval_QID_' + str(qid) + '_' + eval_measure + '.csv'
                    path = os.path.join(eval_path, str(r), split, eval_measure, file_name)

                    if os.path.exists(path):
                        df = pd.read_csv(path)
                        res_all.append(df)

                res_all = pd.concat(res_all)
                res_all = res_all.groupby(['rank', 'group', 'k']).mean().reset_index()
                res_all.to_csv(output_f)
                print("--- Save eval file in ", output_f, " --- \n")


def evaluate(data, query_col, s_attribute, eval_path, args_eval):
    """
    Evaluate the given data by calculating metrics for each query and sensitive attribute.

    :param data : pandas.DataFrame
        The data containing the query identifiers and sensitive attribute.
    :param query_col : str
        The column name for the query identifiers in the data.
    :param s_attribute : str
        The sensitive attribute column name.
    :param eval_path : str
        Path where evaluation results will be saved.
    :param args_eval : dict
        A dictionary containing evaluation configurations, including metrics and other settings.

    :return : None
        This function saves the evaluation results as CSV files.
    """
    if not os.path.exists(eval_path):
        os.makedirs(eval_path)

    qids = data[query_col].unique()

    groups = data[s_attribute].unique()
    for eval_measure in args_eval['metrics']:
        if not os.path.exists(os.path.join(eval_path, eval_measure)):
            os.makedirs(os.path.join(eval_path, eval_measure))
            for qid in qids:
                data_qid = data[data[query_col] == qid]
                res_qid = evaluate_qid(data_qid, eval_measure, s_attribute, groups, args_eval)

                output_f = os.path.join(eval_path, eval_measure, "Eval_QID_" + str(qid) + "_" + eval_measure + ".csv")
                res_qid.to_csv(output_f)
                print("--- Save eval file in ", output_f, " --- \n")


def evaluate_qid(df, eval_measure, s_attribute, sensitive_groups, args_eval):
    """
    Evaluate the data for a single query ID, calculating the specified metrics.

    :param df : pandas.DataFrame
        The data for a single query ID.
    :param eval_measure : str
        The evaluation metric to calculate.
    :param s_attribute : str
        The sensitive attribute column name.
    :param sensitive_groups : list
        A list of sensitive attribute groups.
    :param args_eval : dict
        A dictionary containing evaluation configurations, including rankings and k list.

    :return : pandas.DataFrame
        A DataFrame containing the evaluation results.
    """
    EVAL_RANKINGS = args_eval['rankings']

    seti_quotas = get_quotas_count(df, s_attribute, sensitive_groups=sensitive_groups)

    res_df = pd.DataFrame(columns=["run", "rank", "k", "group", eval_measure])
    k_list = args_eval['k_list']

    for ranking in EVAL_RANKINGS:
        # data sorted by ranking
        ranking_df = get_sort_df(ranking, df, len(df))

        for ki in k_list:
            ki = int(ki)
            res_row = [1, ranking, ki]
            all_row = res_row + ["all"]

            # data sorted by ranking value at top-k
            top_ranking = ranking_df.head(ki)

            if 'individual' in eval_measure:
                if "__" in ranking:
                    yNN = compute_individual_fairness(ranking_df, ranking)
                    all_row.append(yNN)
                else:
                    # not applicable on the output of an LTR model
                    all_row.append(-1)

            if eval_measure == "select_rate":
                # applicable per group
                all_row.append(-1)

            if eval_measure == "diversity":
                # applicable per group
                all_row.append(-1)

            if eval_measure == "exposure":
                # applicable per group
                all_row.append(-1)

            if eval_measure == "igf":
                # applicable per group
                all_row.append(-1)

            res_df.loc[res_df.shape[0]] = all_row

            # group-level evaluation
            cur_quotas = get_quotas_count(top_ranking, s_attribute, sensitive_groups)
            for gi in sensitive_groups:
                gi_row = res_row + [gi]
                if eval_measure == "select_rate":
                    # selection rate to rank inside top-k
                    if gi in cur_quotas and seti_quotas[gi] != 0:
                        if seti_quotas[gi] < 1 / ki:  # at least one candidate in top-k
                            seti_quotas[gi] = 1 / ki
                        gi_row.append(cur_quotas[gi] / seti_quotas[gi])
                    else:
                        gi_row.append(0)

                if eval_measure == "diversity":
                    if gi in cur_quotas:
                        gi_row.append(cur_quotas[gi])
                    else:
                        gi_row.append(0)

                gi_top_ranking = top_ranking[top_ranking[s_attribute] == gi]
                gi_ranking_df = ranking_df[ranking_df[s_attribute] == gi]

                if eval_measure == "exposure":
                    gi_row.append(compute_cumulative_exposer(gi_top_ranking, ki))

                if eval_measure == "igf":
                    if gi_ranking_df is not None:
                        if not gi_top_ranking.shape[0]:
                            gi_row.append(-1)
                        else:
                            gi_row.append(compute_igf_ratio(list(gi_top_ranking["UID"]), gi_ranking_df, ranking))

                if 'individual' in eval_measure:
                    # not applicable to group
                    gi_row.append(-1)

                res_df.loc[res_df.shape[0]] = gi_row

    return res_df


def compute_individual_fairness(data, ranking):
    """
    Compute the individual fairness score for the ranking.

    :param data : pandas.DataFrame
        The data to evaluate fairness.
    :param ranking : str
        The ranking column to use for evaluation.

    :return : float
        The individual fairness score between 0 and 1.
    """
    feature_columns = [col for col in data if 'X' in col and '_' not in col]
    distances_data = pairwise.euclidean_distances(data[feature_columns].to_numpy(),
                                                  data[feature_columns].to_numpy())

    exposers = data[ranking].apply(lambda x: 1 / math.log2(x + 1))
    distances_exposer = pairwise.euclidean_distances(exposers.to_numpy().reshape(-1, 1),
                                                     exposers.to_numpy().reshape(-1, 1))

    yNN = 1 - np.mean(np.abs(distances_data - distances_exposer))
    return yNN


def compute_cumulative_exposer(df_top, ki):
    """
    Compute the cumulative exposure for the top-k items in the ranking.

    :param df_top : pandas.DataFrame
        The top-k items in the ranking.
    :param ki : int
        The number of top-k items.

    :return : float
        The cumulative exposure score.
    """
    if len(df_top) < ki:
        df_top["rank"] = list(range(1, len(df_top) + 1))
    else:
        df_top["rank"] = list(range(1, ki + 1))

    if len(df_top) == 0:
        return -1
    exposer_top_k = sum(df_top['rank'].apply(lambda x: 1 / math.log2(x + 1)))

    return exposer_top_k


def compute_igf_ratio(top_k_IDS, _orig_df, _orig_sort_col):
    """
    Compute the IGF (Item Group Fairness) ratio for the top-k items.

    :param top_k_IDS : list
        A list of IDs representing the top-k items.
    :param _orig_df : pandas.DataFrame
        The original DataFrame containing all items.
    :param _orig_sort_col : str
        The column name to sort the original DataFrame by.

    :return : float
        The IGF ratio.
    """
    accepted_candidates = _orig_df["UID"].isin(top_k_IDS)
    _lowest_accepted_score = min(_orig_df[accepted_candidates][_orig_sort_col])

    rejected_candidates = ~accepted_candidates
    if sum(rejected_candidates) == 0:
        return 1

    _highest_rejected_score = max(_orig_df[rejected_candidates][_orig_sort_col])

    if _highest_rejected_score == 0:
        return 1

    cur_res = min(_orig_df[_orig_df["UID"].isin(top_k_IDS)][_orig_sort_col]) / max(
        _orig_df[~_orig_df["UID"].isin(top_k_IDS)][_orig_sort_col])

    if cur_res > 1:
        return 1
    else:
        return cur_res


def get_quotas_count(_df, s_attribute, sensitive_groups):
    """
    Calculate the quota count for each sensitive group in the data.

    :param _df : pandas.DataFrame
        The data to calculate the quotas for.
    :param s_attribute : str
        The sensitive attribute column name.
    :param sensitive_groups : list
        A list of sensitive attribute groups.

    :return : dict
        A dictionary with the sensitive group names as keys and their corresponding quota counts as values.
    """
    res_dict = {}

    for s in sensitive_groups:
        mask = _df[s_attribute] == s
        res_dict[s] = sum(mask) / len(_df)

    return res_dict


def get_sort_df(_sort_col, _df, _k):
    """
    Sort the DataFrame by the specified column and return the top-k rows.

    :param _sort_col : str
        The column to sort the data by.
    :param _df : pandas.DataFrame
        The DataFrame to be sorted.
    :param _k : int
        The number of top items to return.

    :return : pandas.DataFrame
        The top-k sorted rows from the DataFrame.
    """
    _df[_sort_col] = _df[_sort_col].apply(lambda x: float(x))
    _k = int(_k)
    if "__" in _sort_col:
        sort_df = _df.sort_values(by=_sort_col, ascending=True).head(_k)
    else:
        sort_df = _df.sort_values(by=_sort_col, ascending=False).head(_k)
    return sort_df



import torch
import torch.nn as nn
import numpy as np
from .abstract_model import AbstractBaseModel


r"""
BPR
################################################
Reference:
    Steffen Rendle et al. "BPR: Bayesian Personalized Ranking from Implicit Feedback." in UAI 2009.
"""

class BPR(AbstractBaseModel):
    def __init__(self, config):
        super().__init__(config)
        # self.user_embedding = nn.Embedding(num_users, embedding_dim)
        # self.item_embedding = nn.Embedding(num_items, embedding_dim)
        self.type = ["pair"]
        self.IR_type = ["retrieval", "ranking"]

        self.sigmoid = nn.Sigmoid()
        self.apply(self._init_weights)


    def forward(self, user_dict, item_ids):
        user_embeds = self.get_user_embedding(user_dict)
        item_embeds = self.item_embedding(item_ids)
        dot_product = (user_embeds * item_embeds).sum(1)
        return self.sigmoid(dot_product)

    def get_user_embedding(self, user_dict):
        user = user_dict['user_ids']
        return self.user_embedding(user)

    def compute_loss(self, interaction):
        #user = interaction['user_ids']
        pos_item = interaction['item_ids']
        neg_item = interaction['neg_item_ids']
        pos_score = self.forward(interaction, pos_item)
        neg_score = self.forward(interaction, neg_item)
        return self.bpr_loss(pos_score, neg_score)







import numpy as np
import math
from .Abstract_Reranker import Abstract_Reranker
from tqdm import tqdm,trange

'''
FairRec Model for item-side fair re-ranking

@inproceedings{patro2020fairrec,
  title={Fairrec: Two-sided fairness for personalized recommendations in two-sided platforms},
  author={Patro, Gourab K and Biswas, Arpita and Ganguly, Niloy and Gummadi, Krishna P and Chakraborty, Abhijnan},
  booktitle={Proceedings of the web conference 2020},
  pages={1194--1204},
  year={2020}
}
'''


def greedy_round_robin(m,n,R,l,T,V,U,F):
    # greedy round robin allocation based on a specific ordering of customers (assuming the ordering is done in the relevance scoring matrix before passing it here)

    # creating empty allocations
    B={}
    for u in U:
        B[u]=[]

    # available number of copies of each producer
    Z={} # total availability
    P=range(n) # set of producers
    for p in P:
        Z[p]=l

    # allocating the producers to customers
    for t in range(1,R+1):
        #print("GRR round number==============================",t)
        for i in range(m):
            if T==0:
                return B,F
            u=U[i]
            # choosing the p_ which is available and also in feasible set for the user
            possible=[(Z[p]>0)*(p in F[u])*V[u,p] for p in range(n)]
            p_=np.argmax(possible)

            if (Z[p_]>0) and (p_ in F[u]) and len(F[u])>0:
                B[u].append(p_)
                F[u].remove(p_)
                Z[p_]=Z[p_]-1
                T=T-1
            else:
                return B,F
    # returning the allocation
    return B,F


def FairRec_train(U,P,k,V,alpha, m , n):
    # Allocation set for each customer, initially it is set to empty set
    A={}
    for u in U:
        A[u]=[]

    # feasible set for each customer, initially it is set to P
    F={}
    for u in U:
        F[u]=P[:]
    #print(sum([len(F[u]) for u in U]))

    # number of copies of each producer
    l=int(alpha*m*k/(n+0.0))

    # R= number of rounds of allocation to be done in first GRR
    R=int(math.ceil((l*n)/(m+0.0)))


    # total number of copies to be allocated
    T= l*n

    # first greedy round-robin allocation
    [B,F1]=greedy_round_robin(m,n,R,l,T,V,U[:],F.copy())
    F={}
    F=F1.copy()
    #print("GRR done")
    # adding the allocation
    for u in U:
        A[u]=A[u][:]+B[u][:]

    # second phase
    u_less=[] # customers allocated with <k products till now
    for u in A:
        if len(A[u])<k:
            u_less.append(u)

    # allocating every customer till k products
    for u in u_less:
        scores=V[u,:]
        new=scores.argsort()[-(k+k):][::-1]
        for p in new:
            if p not in A[u]:
                A[u].append(p)
            if len(A[u])==k:
                break

    return A

class FairRec(Abstract_Reranker):
    def __init__(self, config, weights=None):
        super().__init__(config, weights)

    def rerank(self, ranking_score, k, batch_size=64):

        user_size = len(ranking_score)
        U = list(range(user_size))  # list of customers
        P = list(range(self.config['item_num']))
        rerank_list = FairRec_train(U, P, k, ranking_score, alpha=self.config['para'],
                                               m=user_size, n=self.config['item_num'])

        # P=list(range(self.config['item_num'])) # list of producers
        # rerank_list = []
        # for b in trange(int(np.ceil(user_size/batch_size))):
        #     min_index = batch_size * b
        #     max_index = min((b+1) * batch_size, user_size)
        #     U = list(range(max_index-min_index))  # list of customers
        #     batch_rerank_list = FairRec_train(U,P,k, ranking_score[min_index:max_index,:] ,alpha=self.config['para'],
        #                                       m=max_index-min_index, n=self.config['item_num'])
        #     for k in batch_rerank_list.keys():
        #         rerank_list.append(batch_rerank_list[k])
        #print(rerank_list)
        #exit(0)
        return rerank_list

        #rho_reverse = 1/(self.rho*batch_size*self.TopK)





# from recommendation.trainer import RecTrainer
#
# config = {'model': 'BPR', 'data_type': 'pair', 'fair-rank': True, 'rank_model': 'APR', 'use_llm': False, 'log_name': "test", 'dataset': 'steam'}
# trainer = RecTrainer(train_config=config)
# trainer.train()

# from recommendation.reranker import RecReRanker

# config = {'ranking_store_path': 'steam-base-mf', 'model': 'CPFair', 'fair-rank': True, 'log_name': 'test',
#           'fairness_metrics': ["MMF", "GINI"], 'dataset': 'steam'}

# reranker = RecReRanker(train_config=config)
# reranker.rerank()

from search.trainer import SRDTrainer

config = {'model': 'xquad', 'train_config_file': 'Post-processing.yaml', 'dataset': 'clueweb09', 'task': 'search', 'mode': 'train', "log_name": "ClueWeb09-xQuAD", "model_save_dir": "model/", "tmp_dir": "tmp/", "best_model_list": [], "device": "cpu",}
trainer = SRDTrainer(train_config=config)
trainer.train()



import os

import numpy as np
import torch
import torch.nn as nn
from torch.utils.data import Dataset, DataLoader
import random
import json
import os
import ast
from .utils import Build_Adjecent_Matrix, convert_keys_values_to_int


class AbstractDataset(Dataset):
    def __init__(self, df, config):
        """
            Abstract dataset utilized for training

            :param df: The input DataFrame.
            :param config: The merged config file.
        """


        self.config = config
        self.uid_field, self.iid_field, self.pid_field, self.label_field = config['user_id'], config['item_id'], config['group_id'], config['label_id']
        history_column = 'history_behaviors'

        self.M, self.iid2pid = Build_Adjecent_Matrix(config)

        self.user_ids = torch.tensor(df[self.uid_field].values, dtype=torch.long)
        self.item_ids = torch.tensor(df[self.iid_field].values, dtype=torch.long)
        self.label = torch.tensor(df[self.label_field].values, dtype=torch.float32)
        self.group_id = torch.tensor(df[self.pid_field].values, dtype=torch.long)
        self.history_ids = torch.stack([torch.tensor(row,  dtype=torch.long) for row in df[history_column]])
        #self.history_ids = torch.tensor(df[history_column].values, dtype=torch.long)


    def __len__(self):
        return len(self.label)

class PointWiseDataset(AbstractDataset):

    def __getitem__(self, idx):
        """
           getitem function for pointwise dataset

           :param idx: The index of the sample
           :return: The utilized features of the dataset
        """
        return self.user_ids[idx], self.history_ids[idx], self.item_ids[idx], self.group_id[idx], self.label[idx]

class SequentialDataset(AbstractDataset):

    def __init__(self, df, config):
        """
          SequentialDataset initialization:we only utilize the positive label and sample some neg labels

          :param df: The input DataFrame.
          :param config: The merged config file.
       """
        filter_df = df[df[config['label_id']] == 1]
        super().__init__(filter_df, config)

    def __getitem__(self, idx):
        """
           getitem function for sequential dataset

           :param idx: The index of the sample
           :return: The utilized features of the sequential dataset
        """
        return self.user_ids[idx], self.history_ids[idx], self.item_ids[idx], self.group_id[idx]


class PairWiseDataset(AbstractDataset):
    def __init__(self, df, config):
        """
           PairWiseDataset initialization: here we construct the positive samples for each items

           :param df: The input DataFrame.
           :param config: The merged config file.
        """
        filter_df = df[df[config['label_id']] == 1]
        super().__init__(filter_df, config)
        self.user2pos = {}
        for row in filter_df[[self.uid_field, self.iid_field, self.pid_field]].itertuples(index=True):
            #print(row)
            user_id, item_id, pid = row._1, row._2, row._3
            if user_id not in self.user2pos.keys():
                self.user2pos[user_id] = []
            self.user2pos[user_id].append(item_id)
        for key in self.user2pos.keys():
            self.user2pos[key] = list(set(self.user2pos[key]))


    def __getitem__(self, idx):
        """
          getitem function for sequential dataset with negative sampling operations

          :param idx: The index of the sample
          :return: The utilized features of the sequential dataset
       """
        user_id = int(self.user_ids[idx].numpy())
        neg_item = 0
        while neg_item in self.user2pos[user_id]:
            neg_item = random.randint(0, self.config['item_num']-1)

        neg_group = torch.tensor(self.iid2pid[neg_item], dtype=torch.long)
        neg_item = torch.tensor(neg_item, dtype=torch.long)

        return self.user_ids[idx], self.history_ids[idx], self.item_ids[idx], neg_item, self.group_id[idx], neg_group
        #return self.user_ids[idx], self.item_ids[idx], self.label[idx], self.group_id[idx]

class RankingTestDataset(Dataset):
    def __init__(self, df, config):
        with open(os.path.join("recommendation", "processed_dataset", config['dataset'], "iid2pid.json"), "r") as file:
            self.item2pid = json.load(file)
        df['items'] = df['items'].apply(lambda x: np.array(ast.literal_eval(x)))
        self.user_ids = torch.tensor(df["user_id"].values, dtype=torch.long)
        self.items = torch.stack([torch.tensor(row, dtype=torch.long) for row in df["items"]])
        self.pos_length = torch.tensor(df["pos_length"].values, dtype=torch.long)
        self.history_ids = torch.stack([torch.tensor(row, dtype=torch.long) for row in df["history_behaviors"]])

    def __len__(self):
        return len(self.user_ids)

    def __getitem__(self, idx):
        """
             getitem function for RankingDataset dataset under test dataset since in test set we already finish negative samples

             :param idx: The index of the sample
             :return: The utilized features of the RankingDataset
        """
        return self.user_ids[idx], self.history_ids[idx], self.items[idx], self.pos_length[idx]


from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import os

from setuptools import setup, find_packages

install_requires = [
    "numpy>=1.20.3",
    "torch>=1.11.0",
    "PyYAML>=6.0",
    "pandas>=1.2.0",
    "scipy",
    "cvxpy",
    "tqdm>=4.65.0",
    "scikit_learn",
    "pot",
    "mip>=1.15.0",
    "gurobipy>=12.0.1"
]

setup_requires = []


classifiers = ["License :: OSI Approved :: MIT License"]

long_description = (
    "FairDiverse is developed based on Python for "
    "reproducing and developing fairness- and diversity-aware "
    "IR algorithms in a unified, comprehensive framework for "
    "research purpose. In the first version, our library "
    "includes 28 fairness- and diversity algorithms and 16 base models, covering "
    "pre-processing, in-processing and post-processing"
    "for more information: https://github.com/XuChen0427/FairDiverse"
)

# Readthedocs requires Sphinx extensions to be specified as part of
# install_requires in order to build properly.
on_rtd = os.environ.get("READTHEDOCS", None) == "True"
if on_rtd:
    install_requires.extend(setup_requires)

setup(
    name="fairdiverse",
    version="1.0.0",  # please remember to edit recbole/__init__.py in response, once updating the version
    description="A unified, comprehensive fairness- and diversity-aware IR library",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/XuChen0427/FairDiverse",
    author="Chen Xu",
    author_email="chenxu0427ruc@gmail.com",
    #packages=find_packages(),
    packages=[package for package in find_packages() if package.startswith("fairdiverse")],
    include_package_data=True,
    install_requires=install_requires,
    setup_requires=setup_requires,
    extras_require=None,
    zip_safe=False,
    classifiers=classifiers,
    python_requires=">=3.6",
)


import json
import os
import numpy as np
from scipy.sparse import coo_matrix, csr_matrix
import json
import torch


def convert_keys_values_to_int(data):
    """
       Recursively converts dictionary keys and values, as well as string digits, into integers.

       This function will process a dictionary or list and convert all keys and values to integers where applicable.
       If the data is a string that represents a digit, it will be converted to an integer.

       :param data: The input data to be processed, which can be a dictionary, list, or string.
       :return: The input data with keys and values converted to integers if applicable.
   """

    if isinstance(data, dict):
        return {int(k): convert_keys_values_to_int(v) for k, v in data.items()}
    elif isinstance(data, list):
        return [convert_keys_values_to_int(element) for element in data]
    elif isinstance(data, str) and data.isdigit():
        return int(data)
    else:
        return data

def Init_Group_AdjcentMatrix(dataset_name):
    """
        Initializes the group adjacency matrix by loading the processed dataset corresponding to the given dataset name.

        This function checks if the processed dataset exists for the provided dataset name, and if not, raises an error.
        It then loads the `iid2pid.json` file, converts the keys and values to integers, and returns the result.

        :param dataset_name: The name of the dataset to load.
        :return: A dictionary mapping item IDs (iid) to product IDs (pid).
    """


    dir = os.path.join("recommendation", "processed_dataset", dataset_name)
    if not os.path.exists(dir):
        raise ValueError("do not processed such data, please run the ranking phase to generate data for re-ranking")

    with open(os.path.join(dir, "iid2pid.json"), "r") as file:
        iid2pid = json.load(file)
        iid2pid = convert_keys_values_to_int(iid2pid)

    return iid2pid

def get_iid2text(dataset_name):
    """
        Loads and returns the mapping of item IDs (iid) to text descriptions from the processed dataset.

        This function checks if the processed dataset exists for the given dataset name, and if not, raises an error.
        It then loads the `iid2text.json` file, converts the keys and values to integers, and returns the result.

        :param dataset_name: The name of the dataset to load.
        :return: A dictionary mapping item IDs (iid) to their corresponding text descriptions.
    """


    dir = os.path.join("recommendation", "processed_dataset", dataset_name)
    if not os.path.exists(dir):
        raise ValueError("do not processed such data, please run the ranking phase to generate data for re-ranking")

    with open(os.path.join(dir, "iid2text.json"), "r") as file:
        iid2text = json.load(file)
        iid2text = convert_keys_values_to_int(iid2text)

    return iid2text

def Build_Adjecent_Matrix(config):
    """
        Builds an adjacency matrix based on the group-item mapping, initializing it with ones,
        and adjusting rows with no connections.

        This function uses the `Init_Group_AdjcentMatrix` to retrieve a mapping of item IDs (iid) to product IDs (pid),
        constructs an adjacency matrix, and ensures that rows with no connections are assigned a default value.

        :param config: A configuration dictionary containing dataset information and matrix dimensions.
        :return: A tuple containing:
            - A 2D NumPy array representing the adjacency matrix.
            - A dictionary mapping item IDs (iid) to product IDs (pid).
    """


    iid2pid = Init_Group_AdjcentMatrix(config['dataset'])
    row = list(iid2pid.keys())
    col = list(iid2pid.values())
    data = np.ones_like(row)
    M = coo_matrix((data, (row, col)), shape=(config['item_num'], config['group_num']))
    M = M.toarray()

    for i in range(len(M)):
        if np.sum(M[i]) == 0:
            M[i][0] = 1
            iid2pid[i] = 0

    return M, iid2pid

def load_json(file_path):
    """
    Loads a JSON file from the specified file path.

    This function attempts to open and parse a JSON file. If successful, it returns the parsed data.
    If the file is not found, or the file content is not valid JSON, an error message is printed,
    and `None` is returned.

    :param file_path: The path to the JSON file to be loaded.
    :return: The parsed data from the JSON file if successful, otherwise `None`.
    """

    try:
        with open(file_path, 'r', encoding='utf-8') as file:
            data = json.load(file)
        return data
    except FileNotFoundError:
        print(f"Error: The file at {file_path} was not found.")
        return None
    except json.JSONDecodeError:
        print(f"Error: The file at {file_path} is not a valid JSON file.")
        return None
    except Exception as e:
        print(f"An unexpected error occurred: {e}")
        return None



def get_cos_similar_torch(v1, v2, device='cuda'):
    """
        Computes the cosine similarity between two vectors using PyTorch.

        This function calculates the cosine similarity between two input vectors, `v1` and `v2`, using the specified device
        (either `cuda` for GPU or `cpu`). The function utilizes PyTorch's `cosine_similarity` function and returns the result as a NumPy array.

        :param v1: The first vector for cosine similarity calculation.
        :param v2: The second vector for cosine similarity calculation.
        :return: The cosine similarity between `v1` and `v2`.
    """


    import torch.nn.functional as F
    if device == 'cuda':
        v1 = torch.tensor(v1).cuda()
        v2 = torch.tensor(v2).cuda()
        cos_sim = F.cosine_similarity(v1, v2)
        return cos_sim.to(torch.float).cpu().numpy()
    else:
        v1 = torch.tensor(v1).cpu().float()
        v2 = torch.tensor(v2).cpu().float()
        cos_sim = F.cosine_similarity(v1, v2)
        return cos_sim.numpy()




import os.path
import numpy as np
from tqdm import tqdm, trange
from ..utils.group_utils import get_cos_similar_torch
import os
import torch
from transformers import AutoModelForCausalLM, AutoTokenizer


class Grounder(object):
    def __init__(self, config):
        # self.config
        self.grounding_model = config['grounding_model']
        self.llm_path_dict = config['llm_path_dict']
        print(type(self.grounding_model))
        assert self.grounding_model in self.llm_path_dict.keys(), f"Grounding Model {self.grounding_model} Not Found."
        self.grounding_model_path = self.llm_path_dict[self.grounding_model]
        self.ground_in_8bit = config['use_8bit']
        self.saved_embs_filename = os.path.join("recommendation", "llm", "stored_embs", config['saved_embs_filename']) if config['saved_embs_filename'] else None
        self.grounding_batch_size = config['batch_size']
        self.device_map = config['device_map']
        self.device = config['device']

    def load_model_tokenizer(self):
        """
        Loads the tokenizer and model for text generation based on the specified grounding model.

        This method initializes the tokenizer and model required for text generation tasks. It checks if the provided
        grounding model exists within the pre-defined dictionary of model paths. If found, it proceeds to load the tokenizer
        and model using the Hugging Face Transformers library. The tokenizer's padding side is set to 'left', and the model
        is configured with options such as load_in_8bit, torch_dtype, and device_map for efficient memory usage and deployment.
        Additional model configurations are set, including token ID assignments for pad, bos, eos tokens, and enabling
        output of hidden states. If the specified grounding model is not recognized, an exception is raised.

        :param self.grounding_model (str): The name or identifier of the grounding model to be loaded.
        :param self.llm_path_dict (dict): A dictionary mapping model names to their respective local paths.
        :param self.grounding_model_path (str): The path where the model is located, derived from `self.llm_path_dict`.
        :param self.ground_in_8bit (bool): A flag indicating whether to load the model in 8-bit precision for memory efficiency.
        :param self.device_map (str or dict): Specifies how to map model’s tensors to devices; can be a string or a dictionary.

        Note:
        - Ensure that the `self.llm_path_dict` contains the correct paths for each model before calling this method.
        - Model loading can be resource-intensive; ensure sufficient system resources are available.
        """
        if self.grounding_model in self.llm_path_dict.keys():
            self.tokenizer = AutoTokenizer.from_pretrained(self.grounding_model_path)
            self.tokenizer.padding_side = "left"
            self.model = AutoModelForCausalLM.from_pretrained(
                self.grounding_model_path,
                load_in_8bit=self.ground_in_8bit,
                torch_dtype='auto',
                device_map=self.device_map,
            )
        else:
            raise Exception('Illegal Grouding Model Defined.')
        self.model.config.pad_token_id = self.tokenizer.pad_token_id = 0  # unk
        self.model.config.bos_token_id = 1
        self.model.config.eos_token_id = 2
        self.model.config.output_hidden_states = True
        # self.model.half()

    def get_embedding(self, text, index):
        """
        Get the embedding of the input text using the model at a specified hidden layer index.

       :param text : str
            The input text to be embedded.
       :param index : int
            The index of the hidden layer from which to extract the embeddings.

        :return : torch.Tensor
            The averaged embedding vector for the input text.
        """
        tokens = self.tokenizer(text, return_tensors="pt", padding=True, truncation=True, max_length=128).to(self.device)
        input_ids = tokens["input_ids"]
        attention_mask = tokens['attention_mask']
        with torch.no_grad():
            output = self.model(input_ids=input_ids, attention_mask=attention_mask, output_hidden_states=True)
            embedding = output['hidden_states'][index]
            embedding = torch.mean(embedding, dim=1, keepdim=False)
        return embedding

    def get_title_embedding(self, titles, index):
        """
        Get title embeddings using a model's embedding generation capability.

        This method retrieves embeddings for a list of titles. It first checks if a saved embeddings file exists,
        and if so, loads the embeddings from it. If not, it generates the embeddings in batches using the
        `get_embedding` method. The generated or loaded embeddings are then returned as a PyTorch tensor.

        :param titles : List[str]
            A list of strings representing the titles for which embeddings are to be obtained.
        :param index : int
            An index indicating a specific configuration or identifier used in the embedding process.

        :return: torch.Tensor
            A tensor containing the embeddings for the provided titles. Each row corresponds to the embedding of a title.

        Notes
        -----
        - If `saved_embs_filename` is set and the file exists, embeddings are directly loaded from it, bypassing computation.
        - Embeddings are computed in batches to manage memory efficiently, with the batch size defined by `grounding_batch_size`.
        - If embeddings are computed, they can optionally be saved to `saved_embs_filename` for future reuse.
        """
        if self.saved_embs_filename and os.path.exists(self.saved_embs_filename):
            embs = torch.load(self.saved_embs_filename)
        else:
            embs = []
            batch_size = self.grounding_batch_size
            for b in trange(int(np.ceil(len(titles) / batch_size)), desc='Get Title Embedding'):
                min_id = b * batch_size
                max_id = min((b + 1) * batch_size, len(titles))
                text = titles[min_id:max_id]
                emb = self.get_embedding(text, index)
                embs.append(emb)
            embs = torch.cat(embs, dim=0)
            if self.saved_embs_filename:
                torch.save(embs, self.saved_embs_filename)
        return embs

    def map_titles(self, titles, o_emb, index, candidates):
        """
        Map titles to their similarity scores with a given object embedding.

        This method computes the cosine similarity between the embeddings of a list of titles and a given object embedding.
        It processes the titles in batches, calculates their embeddings using `get_embedding` method, and then assesses
        the similarity with the object embedding. The output is a list of top-k similar titles' indices for each set of candidate
        titles provided, along with the corresponding similarity scores.

        :param titles : List[str]
            A list of titles whose embeddings are to be computed and compared.

        :param o_emb : torch.Tensor
            The embedding vector of the object against which the title embeddings will be compared.

        :param index : Any
            The index or data structure used by `get_embedding` to fetch or compute embeddings.

        :param candidates : List[List[int]]
            A list of lists, where each sublist contains indices representing candidate titles for a user.

        :return: Tuple[List[List[int]], List[List[float]]]
            A tuple containing two elements:
            - The first element is a list of lists, where each sublist contains the indices of top-k most similar titles
              for each set of candidates.
            - The second element is a list of lists, where each sublist contains the similarity scores corresponding to the
              top-k titles' indices.

        """
        result = []
        batch_size = self.grounding_batch_size
        t_embs = []
        for b in trange(int(np.ceil(len(titles) / batch_size)), desc='Item Embedding'):
            min_id = b * batch_size
            max_id = min((b + 1) * batch_size, len(titles))
            text = titles[min_id:max_id]
            emb = self.get_embedding(text, index)
            t_embs.append(emb)
        t_embs = torch.cat(t_embs, dim=0)
        scores = []
        for t_emb, cand in tqdm(zip(t_embs, candidates), desc='map titles'):  # Iterate through each user’s predictions.
            cos_sim = get_cos_similar_torch(t_emb, o_emb, device=self.device)  # Calculate the similarity between each prediction and all the titles.
            # print(cos_sim.shape)
            cos_sim = cos_sim[cand]
            # print(cand)
            sorted_elements_with_indices = sorted(zip(cos_sim, cand))[::-1]
            score = [element for element, index in sorted_elements_with_indices]
            topk_list = [index for element, index in sorted_elements_with_indices]
            scores.append(score)
            result.append(topk_list)
        return result, scores

    def get_ranking_itemlist(self, response_result, o_emb, index):
        """
        Processes the response data to extract item candidates and their predicted scores. It utilizes the provided embeddings and index to map and score the candidates, returning a list of ranked items along with their respective scores.

        :param response_result (List[Dict[str, Any]]): A list of dictionaries containing user-specific 'item_candidates' and 'predict' values.
        :param o_emb (np.ndarray): The embedding matrix used for scoring the item candidates.
        :param index (Any): The index object that aids in mapping and retrieving candidate information efficiently.

        :return: Tuple[List[str], List[float]]: A tuple containing two lists:
            - The first list contains the top-ranked item identifiers based on the prediction scores.
            - The second list holds the corresponding prediction scores for each item in the ranked list.
        """
        candidates = [user['item_candidates'] for user in response_result]
        predict = [user['predict'] for user in response_result]
        predict, scores = self.map_titles(predict, o_emb, index, candidates)
        return predict, scores

    def grounding(self, response_results, id2title):
        """
        Grounding method to enhance response results with predicted item rankings and scores.

        This method enhances a list of response results by grounding them with relevant
        item predictions and associated scores. It utilizes a pre-trained model to generate
        embeddings for a provided list of titles and then ranks items within each response
        result based on their similarity to these embeddings.

        :param response_results (List[Dict]): A list of dictionaries, each representing a response result.
        :param id2title (Dict[int, str]): A dictionary mapping IDs to their corresponding title names.

        :return: List[Dict]: The updated list of response results, where each result now includes
          a 'predict_list' key holding the predicted item IDs and a 'scores' key with their
          respective similarity scores.
        """
        # title2id = {v: k for k, v in id2title.items()}
        title_name = list(id2title.values())
        self.load_model_tokenizer()

        title_emb = self.get_title_embedding(title_name, -1)  #Get the embeddings of the last layer for all titles.
        predict, scores = self.get_ranking_itemlist(response_results, title_emb, -1)
        for idx, res in enumerate(response_results):
            res['predict_list'] = predict[idx]
            # res['prediction_list'] = [title2id[i] for i in predict[idx]]
            res['scores'] = scores[idx]
        # print(f'response_result:{response_results}')
        return response_results


import os
import subprocess
from pathlib import Path
import pandas as pd

from .ranker import Ranker
from ..utils.process_tabular_data import norm_features, check_nan, writeToTXT


class RankLib(Ranker):
    """
    Wrapper class to run the available ranking models in the Ranklib library.
    For more information about available models and params check the official documentation: https://sourceforge.net/p/lemur/wiki/RankLib%20How%20to%20use/
    """
    def __init__(self, configs, dataset):
        """
        Initialize the RankLib model with configuration settings and dataset.

        :param configs : dict
            The configuration dictionary that contains hyperparameters and paths needed for training.
        :param dataset : object
            The dataset object containing the necessary columns like feature columns and target variable.
        """
        super().__init__(configs, dataset)

    def train(self, data_train, data_test, run):
        """
        Trains ranking models using RankLib.

        This method generates RankLib-compatible training data and then runs the RankLib training script.

        :param data_train : pandas.DataFrame
            The training dataset to be used for training the ranking model.
        :param data_test : pandas.DataFrame
            The testing dataset to be used for evaluating the ranking model.
        :param run : str
            The identifier for the current training run.
        """
        self.generate_ranklib_data(data_train, data_test, run)
        for experiment in self._get_experiments(run):
            model_path = self.out_dir / str(run) / experiment / self.configs['ranker']
            if not model_path.exists():
                self._run_ranklib_training(experiment, run)

    def predict(self, data, run, file_name):
        """
        Generates predictions using the trained RankLib model.

        This method reads the predictions from the trained model and saves them as a CSV file.

        :param data : pandas.DataFrame
            The dataset on which predictions need to be made.
        :param run : str
            The identifier for the current run.
        :param file_name : str
            The file name to save the predictions as a CSV.

        :return : pandas.DataFrame
            A DataFrame containing the predictions.
        """
        predictions = self.read_predictions(data, run)
        pred_dir_path = os.path.join(self.out_dir, str(run), "predictions")
        os.makedirs(pred_dir_path, exist_ok=True)
        predictions.to_csv(os.path.join(pred_dir_path, f'{file_name}_pred.csv'))
        return predictions

    def _get_experiments(self, run):
        """
        Fetches all experiment directories that contain '__' in the name.

        :param run : str
            The identifier for the run.

        :return : list
            A list of experiment directory names.
        """
        return [f for f in os.listdir(self.out_dir / str(run)) if "__" in f]

    def _run_ranklib_training(self, experiment, run):
        """
        Executes the RankLib training script.

        This method calls an external script to train the ranking model with the provided configurations.

        :param experiment : str
            The name of the current experiment.
        :param run : str
            The identifier for the current run.
        """
        project_dir = Path.cwd()
        try:
            subprocess.check_call([
                str(project_dir / self.ranker_path / "run-LTR-model.sh"),
                str(project_dir / self.ranker_path),
                str(self.configs['metric']),
                str(self.configs['top_k']),
                str(self.configs['rel_max']),
                str(self.configs['ranker']),
                str(self.configs['ranker_id']),
                str(self.out_dir / str(run) / experiment),
                str(self.configs['lr']),
                str(self.configs['epochs']),
                "none"
            ])
        except subprocess.CalledProcessError as e:
            print(f"Error occurred: {e}")
            print(f"Command output: {e.output}")

    def generate_ranklib_data(self, data_train, data_test, run):
        """
        Generates data formatted for RankLib training and testing.

        This method prepares the training and testing data for RankLib by generating the required feature matrix
        and label information in a format that RankLib can process.

        :param data_train : pandas.DataFrame
            The training dataset.
        :param data_test : pandas.DataFrame
            The testing dataset.
        :param run : str
            The identifier for the current run.
        """
        experiments = list(zip(self.configs['train_data'], self.configs['test_data']))

        for train_name, test_name in experiments:
            data_train_copy = data_train.copy()
            data_test_copy = data_test.copy()

            out_dir = self.out_dir / str(run) / f"{train_name}__{test_name}"
            if not os.path.exists(out_dir):
                out_dir.mkdir(parents=True, exist_ok=True)

                cols_train = self._get_feature_columns(train_name)
                cols_test = self._get_feature_columns(test_name)

                self._validate_data(data_train_copy, cols_train, "train")
                self._validate_data(data_test_copy, cols_test, "test")

                data_train_copy = self._process_data(data_train_copy, cols_train)
                data_test_copy = self._process_data(data_test_copy, cols_test)

                self.create_ranklib_data(cols_train, data_train_copy, out_dir, "train")
                self.create_ranklib_data(cols_test, data_test_copy, out_dir, "test")

    def _get_feature_columns(self, mode):
        """
        Returns feature columns with fairness suffix if applicable.

        :param mode : str
            The mode can be "fair" or another value. If "fair", fairness-related features will be included.

        :return : list
            A list of feature column names.
        """
        suffix = "_fair" if mode == "fair" else ""
        return [col + suffix for col in self.dataset.feature_cols] + [self.dataset.score_col + suffix]

    def _validate_data(self, data, cols, split):
        """
        Checks for NaN values in the dataset.

        This method ensures that there are no missing values in the data before further processing.

        :param data : pandas.DataFrame
            The dataset to be validated.
        :param cols : list
            A list of column names to check for NaN values.
        :param split : str
            The split type, either "train" or "test".

        :raises ValueError : if NaN values are detected.
        """
        if check_nan(data, cols):
            raise ValueError(f"NaN values detected in {split} data!")

    def _process_data(self, data, cols):
        """
        Assigns judgement scores and normalizes features.

        This method applies judgement scoring based on relevance and normalizes the features for RankLib.

        :param data : pandas.DataFrame
            The dataset to be processed.
        :param cols : list
            The list of feature column names.

        :return : pandas.DataFrame
            The processed dataset with normalized features.
        """
        data = data.groupby(self.dataset.query_col).apply(
            lambda x: self.assign_judgement(x, self.dataset.pos_th, cols)).reset_index(drop=True)
        return norm_features(cols, data)

    def create_ranklib_data(self, cols, data, out_dir, split):
        """
        Formats and writes data for RankLib.

        This method prepares the data by formatting it according to RankLib's required format and writes it
        to a text file.

        :param cols : list
            The list of feature columns.
        :param data : pandas.DataFrame
            The data to be written to a text file.
        :param out_dir : Path
            The output directory where the file will be saved.
        :param split : str
            The type of data, either "train" or "test".
        """
        data = self._format_ranklib_data(cols, data)
        output_f = out_dir / f"{split}_ranklib.txt"
        writeToTXT(output_f, data)

    def _format_ranklib_data(self, cols, data):
        """
        Formats features for RankLib.

        This method ensures that the features are formatted in a way RankLib can consume, and shuffles
        the data within query groups.

        :param cols : list
            The list of feature columns.
        :param data : pandas.DataFrame
            The dataset to be formatted.

        :return : pandas.DataFrame
            The formatted dataset ready for RankLib.
        """
        for idx, col in enumerate(cols):
            data[col] = data[col].apply(lambda x: f"{idx + 1}:{round(x, 4)}")

        data["QID"] = data["QID"].apply(lambda x: f"qid:{x}")
        data["UID"] = data[["UID", "judgement", cols[-1]]].astype(str).apply(
            lambda x: f"#docid={x.iloc[0]};rel={x.iloc[1]};{cols[-1]}={x.iloc[2]};",
            axis=1
        )

        # Shuffle within query groups
        data = pd.concat([g.sample(frac=1) for _, g in data.groupby("QID")]).reset_index(drop=True)

        return data[["judgement", "QID"] + cols + ["UID"]]

    def assign_judgement(self, x, th, cols):
        """
        Assigns judgement scores based on relevance ranking.

        This method assigns a judgement score to each document based on its relevance to a query.

        :param x : pandas.DataFrame
            The subset of data belonging to a single query.
        :param th : float
            The threshold for classifying relevance.
        :param cols : list
            The list of feature columns.

        :return : pandas.DataFrame
            The data with the assigned judgement scores.
        """
        mask_pos = x[self.dataset.score_col].apply(lambda x: round(x, 2) > th)
        pos_x, neg_x = x[mask_pos], x[~mask_pos]

        pos_x['judgement'] = pos_x[cols[-1]].rank(ascending=True, method='dense')
        pos_x['judgement'] = self._scale_judgements(pos_x)

        neg_x['judgement'] = 0
        return pd.concat([pos_x, neg_x])

    def _scale_judgements(self, pos_x):
        """
        Scales and rounds judgement values.

        This method scales the judgement values to a defined range and ensures all values are within the valid range.

        :param pos_x : pandas.DataFrame
            The data with the initial judgement scores.

        :return : pandas.Series
            The scaled judgement scores.
        """
        min_rank, max_rank = pos_x['judgement'].min(), pos_x['judgement'].max()
        pos_x['judgement'] = ((pos_x['judgement'] - min_rank) / (max_rank - min_rank + 1)) * self.configs['rel_max']
        pos_x['judgement'] = pos_x['judgement'].round().astype(int)

        # Ensure values are within valid range
        if pos_x['judgement'].max() < self.configs['rel_max']:
            pos_x['judgement'] += self.configs['rel_max'] - pos_x['judgement'].max()
        pos_x['judgement'] = pos_x['judgement'].clip(lower=1)

        return pos_x['judgement']

    def read_predictions(self, data, run):
        """
        Retrieves LTR predictions for the dataset.

        This method loads the predictions from the trained RankLib model.

        :param data : pandas.DataFrame
            The dataset for which predictions need to be made.
        :param run : str
            The identifier for the run.

        :return : pandas.DataFrame
            The dataset with predictions added.
        """
        pred_dir = self.out_dir / str(run)
        predictions = get_LTR_predict(data, pred_dir, self.configs['ranker'], self.dataset.score_col,
                                      self.dataset.query_col, self.dataset.ID)
        return predictions


def get_LTR_predict(data, out_dir, ranker, score_col, query_col, id_col):
    """
    Fetches RankLib prediction scores.

    This method loads prediction scores from the model and merges them with the provided dataset.

    :param data : pandas.DataFrame
        The dataset that needs the predictions.
    :param out_dir : Path
        The directory where the RankLib predictions are stored.
    :param ranker : str
        The name of the ranking model used.
    :param score_col : str
        The column name of the score in the dataset.
    :param query_col : str
        The column representing queries.
    :param id_col : str
        The unique identifier for each data point.

    :return : pandas.DataFrame
        The dataset with added prediction scores.
    """


    experiments = [f for f in os.listdir(out_dir) if "__" in f]

    for experiment in experiments:
        train_set, test_set = experiment.split("__")

        train_set = f"_{train_set}" if train_set == "fair" else ""
        test_set = f"_{test_set}" if test_set == "fair" else ""

        pred_col = f"{score_col}{train_set}__{score_col}{test_set}"

        score_pred = get_prediction_scores(out_dir / experiment / ranker)
        data = data[data[id_col].astype(str).isin(score_pred)]
        data[pred_col] = data[id_col].apply(lambda x: score_pred.get(str(x), 0))
    return data


def get_prediction_scores(pred_path):
    """
    Retrieves prediction scores from the latest RankLib experiment.

    This method reads the predictions generated from the latest experiment and returns them.

    :param pred_path : Path
        The directory containing the prediction files.

    :return : dict
        A dictionary mapping document IDs to predicted scores.
    """


    sub_experiments = [x for x in os.listdir(pred_path) if "experiments_" in x]
    if not sub_experiments:
        raise ValueError(f"No predictions found in {pred_path}!")

    latest_exp = max(sub_experiments, key=lambda x: os.path.getmtime(os.path.join(pred_path, x)))
    pred_file = Path(pred_path) / latest_exp / "predictions" / "prediction.txt"

    if pred_file.exists():
        print(f"**** Reading predictions from {pred_file}")
        with pred_file.open("r") as file:
            lines = file.read().splitlines()
            return {li.split(" ")[2].split(";")[0].replace("docid=", ""): int(li.split(" ")[3]) for li in lines}

    raise ValueError(f"Prediction file not found in {pred_path}!")



import os
import numpy as np
from .Abstract_Reranker import Abstract_Reranker
from tqdm import tqdm,trange
import cvxpy as cp
import math

r"""
P-MMF: an online algorithm to support the worst-off groups. P-MMF is awarded as WWW spotlight best-paper-candidate

@inproceedings{xu2023p,
  title={P-MMF: Provider max-min fairness re-ranking in recommender system},
  author={Xu, Chen and Chen, Sirui and Xu, Jun and Shen, Weiran and Zhang, Xiao and Wang, Gang and Dong, Zhenhua},
  booktitle={Proceedings of the ACM Web Conference 2023},
  pages={3701--3711},
  year={2023}
}

"""

def compute_projection_maxmin_fairness_with_order(ordered_tilde_dual, rho, lambd):

    m = len(rho)
    answer = cp.Variable(m)
    objective = cp.Minimize(cp.sum_squares(cp.multiply(rho,answer) - cp.multiply(rho, ordered_tilde_dual)))
    #objective = cp.Minimize(cp.sum(cp.multiply(rho,answer) - cp.multiply(rho, ordered_tilde_dual)))
    constraints = []
    for i in range(1, m+1):
        constraints += [cp.sum(cp.multiply(rho[:i],answer[:i])) >= -lambd]
    prob = cp.Problem(objective, constraints)
    prob.solve()
    #print(type(result))
    #exit(0)
    #print(type(answer.value))
    return answer.value



def compute_next_dual(eta, rho, dual, gradient, lambd):
    #rho = self.problem.data.rho
    tilde_dual = dual - eta*gradient/rho/rho
    order = np.argsort(tilde_dual*rho)
    ordered_tilde_dual = tilde_dual[order]
    # print ordered_tilde_dual*rho
    ordered_next_dual = compute_projection_maxmin_fairness_with_order(ordered_tilde_dual, rho[order], lambd)
    # print(ordered_next_dual)
    # print("tilde_dual", rho*tilde_dual)
    # print("next_dual", rho*ordered_next_dual[order.argsort()])
    return ordered_next_dual[order.argsort()]

class PMMF(Abstract_Reranker):
    def __init__(self, config, weights = None):
        super().__init__(config, weights)


    def rerank(self, ranking_score, k):
        ## its parameters

        lambd = self.config['lambda']
        learning_rate = self.config['learning_rate']
        gamma = self.config['gamma']


        user_size = len(ranking_score)
        assert len(self.weights) == self.config['group_num']
        B_t = user_size * k * self.weights

        #B_l = np.zeros(self.group_num)
        rerank_list = []

        mu_t = np.zeros(self.config['group_num'])
        eta = learning_rate / math.sqrt(self.config['item_num'])
        gradient_cusum = np.zeros(self.config['group_num'])

        for u in trange(user_size):
            x_title = ranking_score[u, :] - np.matmul(self.M, mu_t)
            mask = np.matmul(self.M, (B_t > 0).astype(np.float32))
            mask = (1.0 - mask) * -10000.0
            x = np.argsort(x_title + mask, axis=-1)[::-1]
            x_allocation = x[:k]
            re_allocation = np.argsort(ranking_score[u, x_allocation])[::-1]
            x_allocation = x_allocation[re_allocation]
            rerank_list.append(x_allocation)
            B_t = B_t - np.sum(self.M[x_allocation], axis=0, keepdims=False)
            gradient = -np.mean(self.M[x_allocation], axis=0, keepdims=False) + self.weights

            # gradient_list.append(gradient)
            gradient = gamma * gradient + (1 - gamma) * gradient_cusum
            gradient_cusum = gradient
            # gradient = -(B_0-B_t)/((t+1)*K) + rho
            for g in range(1):
                mu_t = compute_next_dual(eta, self.weights, mu_t, gradient, lambd)

        return rerank_list




import torch
import torch.nn as nn
import numpy as np
from .abstract_model import AbstractBaseModel


r"""
MF
################################################
Reference:
    Matrix Factorization Models for learned embeddings of users and items
"""

class MF(AbstractBaseModel):
    def __init__(self, config):
        super().__init__(config)
        # self.user_embedding = nn.Embedding(num_users, embedding_dim)
        # self.item_embedding = nn.Embedding(num_items, embedding_dim)
        self.type = ["point"]
        self.IR_type = ["retrieval", "ranking"]

        self.sigmoid = nn.Sigmoid()
        self.loss = nn.BCELoss(reduction='none')
        self.apply(self._init_weights)


    def forward(self, user_dict, item_ids):
        user = user_dict['user_ids']
        user_embeds = self.user_embedding(user)
        item_embeds = self.item_embedding(item_ids)
        dot_product = (user_embeds * item_embeds).sum(1)
        return self.sigmoid(dot_product)

    def get_user_embedding(self, user_dict):
        user = user_dict['user_ids']
        return self.user_embedding(user)

    def compute_loss(self, interaction):

        #user = interaction['user_ids']
        item = interaction['item_ids']
        label = interaction['label']
        output = self.forward(interaction, item)
        return self.loss(output, label)




